"""
pgxnclient -- zip file utilities
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os
import stat
import zipfile

from pgxnclient.i18n import _
from pgxnclient.errors import PgxnClientException
from pgxnclient.archive import Archive

import logging

logger = logging.getLogger('pgxnclient.zip')


class ZipArchive(Archive):
    """Handle .zip archives"""

    _file = None

    def can_open(self):
        return zipfile.is_zipfile(self.filename)

    def open(self):
        assert not self._file, "archive already open"
        try:
            self._file = zipfile.ZipFile(self.filename, 'r')
        except Exception as e:
            raise PgxnClientException(
                _("cannot open archive '%s': %s") % (self.filename, e)
            )

    def close(self):
        if self._file is not None:
            self._file.close()
            self._file = None

    def list_files(self):
        assert self._file, "archive not open"
        return self._file.namelist()

    def read(self, fn):
        assert self._file, "archive not open"
        return self._file.read(fn)

    def unpack(self, destdir):
        zipname = self.filename
        logger.info(_("unpacking: %s"), zipname)
        destdir = os.path.abspath(destdir)
        self.open()
        try:
            for fn in self.list_files():
                fname = os.path.abspath(os.path.join(destdir, fn))
                if not fname.startswith(destdir):
                    raise PgxnClientException(
                        _("archive file '%s' trying to escape!") % fname
                    )

                # Looks like checking for a trailing / is the only way to
                # tell if the file is a directory.
                if fn.endswith('/'):
                    os.makedirs(fname)
                    continue

                # The directory is not always explicitly present in the archive
                if not os.path.exists(os.path.dirname(fname)):
                    os.makedirs(os.path.dirname(fname))

                # Copy the file content
                logger.debug(_("saving: %s"), fname)
                fout = open(fname, "wb")
                try:
                    data = self.read(fn)
                    # In order to restore the executable bit, I haven't find
                    # anything that looks like an executable flag in the zipinfo,
                    # so look at the hashbangs...
                    isexec = data[:2] == b'#!'
                    fout.write(data)
                finally:
                    fout.close()

                if isexec:
                    os.chmod(
                        fname, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC
                    )

        finally:
            self.close()

        return self._find_work_directory(destdir)


def unpack(filename, destdir):
    return ZipArchive(filename).unpack(destdir)



"""
pgxnclient -- main package
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os

from pgxnclient.spec import Spec
from pgxnclient.utils.semver import SemVer
from pgxnclient.utils.strings import Label, Term, Identifier


__version__ = '1.3.1'

# Paths where to find the command executables.
# If relative, it's from the `pgxnclient` package directory.
# Distribution packagers may move them around if they wish.
#
# Only one of the paths should be marked as "public": it will be returned by
# pgxn help --libexec
LIBEXECDIRS = [
    # public, path
    (False, './libexec/'),
    (True, '/usr/local/libexec/pgxnclient/'),
]


assert (
    len([x for x in LIBEXECDIRS if x[0]]) == 1
), "only one libexec directory should be public"

__all__ = [
    'Spec',
    'SemVer',
    'Label',
    'Term',
    'Identifier',
    'get_scripts_dirs',
    'get_public_script_dir',
    'find_script',
]


def get_scripts_dirs():
    """
    Return the absolute path of the directories containing the client scripts.
    """
    return [
        os.path.normpath(os.path.join(os.path.dirname(__file__), p))
        for (_, p) in LIBEXECDIRS
    ]


def get_public_scripts_dir():
    """
    Return the absolute path of the public directory for the client scripts.
    """
    return [
        os.path.normpath(os.path.join(os.path.dirname(__file__), p))
        for (public, p) in LIBEXECDIRS
        if public
    ][0]


def find_script(name):
    """Return the absoulute path of a pgxn script.

    The script are usually found in the `LIBEXEC` dir, but any script on the
    path will do (they are usually prefixed by ``pgxn-``).

    Return `None` if the script is not found.
    """
    path = os.environ.get('PATH', '').split(os.pathsep)
    path[0:0] = get_scripts_dirs()
    for p in path:
        fn = os.path.join(p, name)
        if os.path.isfile(fn):
            return fn



import unittest

from pgxnclient import Label, Term, Identifier


class LabelTestCase(unittest.TestCase):
    def test_ok(self):
        for s in [
            'd',
            'a1234',
            'abcd1234-5432XYZ',
            'a12345678901234567890123456789012345678901234567890123456789012',
        ]:
            self.assertEqual(Label(s), s)
            self.assertEqual(Label(s), Label(s))
            self.assert_(Label(s) <= Label(s))
            self.assert_(Label(s) >= Label(s))

    def test_bad(self):
        def ar(s):
            try:
                Label(s)
            except ValueError:
                pass
            else:
                self.fail("ValueError not raised: '%s'" % s)

        for s in [
            '',
            ' a',
            'a ',
            '1a',
            '-a',
            'a-',
            'a_b',
            'a123456789012345678901234567890123456789012345678901234567890123',
        ]:
            ar(s)

    def test_compare(self):
        self.assertEqual(Label('a'), Label('A'))
        self.assertNotEqual(str(Label('a')), str(Label('A')))  # preserving

    def test_order(self):
        self.assert_(Label('a') < Label('B') < Label('c'))
        self.assert_(Label('A') < Label('b') < Label('C'))
        self.assert_(Label('a') <= Label('B') <= Label('c'))
        self.assert_(Label('A') <= Label('b') <= Label('C'))
        self.assert_(Label('c') > Label('B') > Label('a'))
        self.assert_(Label('C') > Label('b') > Label('A'))
        self.assert_(Label('c') >= Label('B') >= Label('a'))
        self.assert_(Label('C') >= Label('b') >= Label('A'))


class TermTestCase(unittest.TestCase):
    def test_ok(self):
        for s in ['aa' 'adfkjh"()']:
            self.assertEqual(Term(s), s)
            self.assertEqual(Term(s), Term(s))
            self.assert_(Term(s) <= Term(s))
            self.assert_(Term(s) >= Term(s))

    def test_bad(self):
        def ar(s):
            try:
                Term(s)
            except ValueError:
                pass
            else:
                self.fail("ValueError not raised: '%s'" % s)

        for s in ['a', 'aa ', 'a/a', 'a\\a', 'a\ta', 'aa\x01']:
            ar(s)


class TestIdentifier(unittest.TestCase):
    def test_nonblank(self):
        self.assertRaises(ValueError, Identifier, "")

    def test_unquoted(self):
        for s in ['x', 'xxxxx', 'abcxyz_0189', 'ABCXYZ_0189']:
            self.assertEqual(Identifier(s), s)

    def test_quoted(self):
        for s, q in [('x-y', '"x-y"'), (' ', '" "'), ('x"y', '"x""y"')]:
            self.assertEqual(Identifier(s), q)


if __name__ == '__main__':
    unittest.main()



"""
pgxnclient -- test suite package

The test suite can be run via setup.py test. But you better use "make check"
in order to correctly set up the pythonpath.
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client


import unittest


# fix unittest maintainers stubborness: see Python issue #9424
if unittest.TestCase.assert_ is not unittest.TestCase.assertTrue:
    # Vaffanculo, Wolf
    unittest.TestCase.assert_ = unittest.TestCase.assertTrue
    unittest.TestCase.assertEquals = unittest.TestCase.assertEqual


if __name__ == '__main__':
    unittest.main()



"""
pgxnclient -- package exceptions

These exceptions can be used to signal expected problems and to exit in a
controlled way from the program.

"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client


class PgxnException(Exception):
    """Base class for the exceptions known in the pgxn package."""


class PgxnClientException(PgxnException):
    """Base class for the exceptions raised by the pgxnclient package."""


class UserAbort(PgxnClientException):
    """The user requested to stop the operation."""


class BadSpecError(PgxnClientException):
    """A bad package specification."""


class ProcessError(PgxnClientException):
    """An error raised calling an external program."""


class InsufficientPrivileges(PgxnClientException):
    """Operation will fail because the user is too lame."""


class NotFound(PgxnException):
    """Something requested by the user not found on PGXN"""


class NetworkError(PgxnClientException):
    """An error from the other side of the wire."""


class BadChecksum(PgxnClientException):
    """A downloaded file is not what expected."""


class ResourceNotFound(NetworkError):
    """Resource not found on the server."""


class BadRequestError(Exception):
    """Bad request from our side.

    This exception is a basic one because it should be rased upon an error
    on our side.
    """



import os
import shutil
import tempfile
import unittest

from mock import patch, Mock
from six.moves.urllib.parse import quote

from pgxnclient.tar import TarArchive
from pgxnclient.zip import ZipArchive
from pgxnclient.errors import (
    PgxnClientException,
    ResourceNotFound,
    InsufficientPrivileges,
)
from .testutils import ifunlink, get_test_filename


class FakeFile(object):
    def __init__(self, *args):
        self._f = open(*args)
        self.url = None

    def __enter__(self):
        self._f.__enter__()
        return self

    def __exit__(self, type, value, traceback):
        self._f.__exit__(type, value, traceback)

    def __getattr__(self, attr):
        return getattr(self._f, attr)


def fake_get_file(url, urlmap=None):
    if urlmap:
        url = urlmap.get(url, url)
    fn = get_test_filename(quote(url, safe=""))
    if not os.path.exists(fn):
        raise ResourceNotFound(fn)
    f = FakeFile(fn, 'rb')
    f.url = url
    return f


def fake_pg_config(**map):
    def f(what):
        return map[what]

    return f


class InfoTestCase(unittest.TestCase):
    def _get_output(self, cmdline):
        @patch('sys.stdout')
        @patch('pgxnclient.network.get_file')
        def do(mock, stdout):
            stdout.encoding = 'UTF-8'
            mock.side_effect = fake_get_file

            from pgxnclient.cli import main

            main(cmdline)
            return get_stdout_data(stdout)

        return do()

    def test_info(self):
        output = self._get_output(['info', '--versions', 'foobar'])
        self.assertEqual(
            output,
            b"""\
foobar 0.43.2b1 testing
foobar 0.42.1 stable
foobar 0.42.0 stable
""",
        )

    def test_info_op(self):
        output = self._get_output(['info', '--versions', 'foobar>0.42.0'])
        self.assertEqual(
            output,
            b"""\
foobar 0.43.2b1 testing
foobar 0.42.1 stable
""",
        )

    def test_info_empty(self):
        output = self._get_output(['info', '--versions', 'foobar>=0.43.2'])
        self.assertEqual(output, b"")

    def test_info_case_insensitive(self):
        output = self._get_output(['info', '--versions', 'Foobar'])
        self.assertEqual(
            output,
            b"""\
foobar 0.43.2b1 testing
foobar 0.42.1 stable
foobar 0.42.0 stable
""",
        )

    def test_mirrors_list(self):
        output = self._get_output(['mirror'])
        self.assertEqual(
            output,
            b"""\
http://pgxn.depesz.com/
http://www.postgres-support.ch/pgxn/
http://pgxn.justatheory.com/
http://pgxn.darkixion.com/
http://mirrors.cat.pdx.edu/pgxn/
http://pgxn.dalibo.org/
http://pgxn.cxsoftware.org/
http://api.pgxn.org/
""",
        )

    def test_mirror_info(self):
        output = self._get_output(['mirror', 'http://pgxn.justatheory.com/'])
        self.assertEqual(
            output,
            b"""\
uri: http://pgxn.justatheory.com/
frequency: daily
location: Portland, OR, USA
bandwidth: Cable
organization: David E. Wheeler
email: justatheory.com|pgxn
timezone: America/Los_Angeles
src: rsync://master.pgxn.org/pgxn/
rsync: 
notes: 

""",  # noqa
        )

    @patch('sys.stdout')
    @patch('pgxnclient.network.get_file')
    def test_nonascii(self, mock_get, stdout):
        mock_get.side_effect = fake_get_file

        from pgxnclient.cli import main

        stdout.encoding = 'UTF-8'
        main(['info', 'first_last_agg'])
        out = max(get_stdout_data(stdout))
        if not isinstance(out, int):
            out = ord(out)
        assert out > 127

        stdout.reset_mock()
        stdout.encoding = None
        main(['info', 'first_last_agg'])
        out = max(get_stdout_data(stdout))
        if not isinstance(out, int):
            out = ord(out)
        assert out < 127


class CommandTestCase(unittest.TestCase):
    def test_popen_raises(self):
        from pgxnclient.commands import Command

        c = Command([])
        self.assertRaises(
            PgxnClientException, c.popen, "this-script-doesnt-exist"
        )

    @patch('sys.stdout')
    def test_empty_command(self, stdout):
        from pgxnclient.cli import main

        stdout.encoding = None
        main([])
        out = get_stdout_data(stdout)

        # The output is the same of 'help'
        stdout.reset_mock()
        main(["help"])
        out_help = get_stdout_data(stdout)
        assert out == out_help


class DownloadTestCase(unittest.TestCase):
    @patch('pgxnclient.network.get_file')
    def test_download_latest(self, mock):
        mock.side_effect = fake_get_file

        fn = 'foobar-0.42.1.zip'
        self.assert_(not os.path.exists(fn))

        from pgxnclient.cli import main

        try:
            main(['download', 'foobar'])
            self.assert_(os.path.exists(fn))
        finally:
            ifunlink(fn)

    @patch('pgxnclient.network.get_file')
    def test_download_testing(self, mock):
        mock.side_effect = fake_get_file

        fn = 'foobar-0.43.2b1.zip'
        self.assert_(not os.path.exists(fn))

        from pgxnclient.cli import main

        try:
            main(['download', '--testing', 'foobar'])
            self.assert_(os.path.exists(fn))
        finally:
            ifunlink(fn)

    @patch('pgxnclient.network.get_file')
    def test_download_url(self, mock):
        mock.side_effect = fake_get_file

        fn = 'foobar-0.43.2b1.zip'
        self.assert_(not os.path.exists(fn))

        from pgxnclient.cli import main

        try:
            main(
                [
                    'download',
                    'https://api.pgxn.org/dist/foobar/0.43.2b1/foobar-0.43.2b1.zip',
                ]
            )
            self.assert_(os.path.exists(fn))
        finally:
            ifunlink(fn)

    @patch('pgxnclient.network.get_file')
    def test_download_ext(self, mock):
        mock.side_effect = fake_get_file

        fn = 'pg_amqp-0.3.0.zip'
        self.assert_(not os.path.exists(fn))

        from pgxnclient.cli import main

        try:
            main(['download', 'amqp'])
            self.assert_(os.path.exists(fn))
        finally:
            ifunlink(fn)

    @patch('pgxnclient.network.get_file')
    def test_download_rename(self, mock):
        mock.side_effect = fake_get_file

        fn = 'foobar-0.42.1.zip'
        fn1 = 'foobar-0.42.1-1.zip'
        fn2 = 'foobar-0.42.1-2.zip'

        for tmp in (fn, fn1, fn2):
            self.assert_(not os.path.exists(tmp))

        try:
            f = open(fn, "w")
            f.write('test')
            f.close()

            from pgxnclient.cli import main

            main(['download', 'foobar'])
            self.assert_(os.path.exists(fn1))
            self.assert_(not os.path.exists(fn2))

            main(['download', 'foobar'])
            self.assert_(os.path.exists(fn2))

            f = open(fn)
            self.assertEquals(f.read(), 'test')
            f.close()

        finally:
            ifunlink(fn)
            ifunlink(fn1)
            ifunlink(fn2)

    @patch('pgxnclient.network.get_file')
    def test_download_bad_sha1(self, mock):
        def fakefake(url):
            return fake_get_file(
                url,
                urlmap={
                    'https://api.pgxn.org/dist/foobar/0.42.1/META.json': 'https://api.pgxn.org/dist/foobar/0.42.1/META-badsha1.json'  # noqa
                },
            )

        mock.side_effect = fakefake

        fn = 'foobar-0.42.1.zip'
        self.assert_(not os.path.exists(fn))

        try:
            from pgxnclient.cli import main
            from pgxnclient.errors import BadChecksum

            self.assertRaises(BadChecksum, main, ['download', 'foobar'])

            self.assert_(not os.path.exists(fn))

        finally:
            ifunlink(fn)

    @patch('pgxnclient.network.get_file')
    def test_download_case_insensitive(self, mock):
        mock.side_effect = fake_get_file

        fn = 'pyrseas-0.4.1.zip'
        self.assert_(not os.path.exists(fn))

        from pgxnclient.cli import main

        try:
            main(['download', 'pyrseas'])
            self.assert_(os.path.exists(fn))
        finally:
            ifunlink(fn)

        try:
            main(['download', 'Pyrseas'])
            self.assert_(os.path.exists(fn))
        finally:
            ifunlink(fn)

    def test_version(self):
        from pgxnclient import Spec
        from pgxnclient.commands.install import Download
        from pgxnclient.errors import ResourceNotFound

        opt = Mock()
        opt.status = Spec.STABLE
        cmd = Download(opt)

        for spec, res, data in [
            ('foo', '1.2.0', {'stable': ['1.2.0']}),
            ('foo', '1.2.0', {'stable': ['1.2.0', '1.2.0b']}),
            ('foo=1.2', '1.2.0', {'stable': ['1.2.0']}),
            ('foo>=1.1', '1.2.0', {'stable': ['1.1.0', '1.2.0']}),
            (
                'foo>=1.1',
                '1.2.0',
                {
                    'stable': ['1.1.0', '1.2.0'],
                    'testing': ['1.3.0'],
                    'unstable': ['1.4.0'],
                },
            ),
        ]:
            spec = Spec.parse(spec)
            data = {
                'releases': dict(
                    [
                        (k, [{'version': v} for v in vs])
                        for k, vs in data.items()
                    ]
                )
            }

            self.assertEqual(res, cmd.get_best_version(data, spec))

        for spec, res, data in [
            ('foo>=1.3', '1.2.0', {'stable': ['1.2.0']}),
            ('foo>=1.3', '1.2.0', {'stable': ['1.2.0'], 'testing': ['1.3.0']}),
        ]:
            spec = Spec.parse(spec)
            data = {
                'releases': dict(
                    [
                        (k, [{'version': v} for v in vs])
                        for k, vs in data.items()
                    ]
                )
            }

            self.assertRaises(
                ResourceNotFound, cmd.get_best_version, data, spec
            )

        opt.status = Spec.TESTING

        for spec, res, data in [
            (
                'foo>=1.1',
                '1.3.0',
                {
                    'stable': ['1.1.0', '1.2.0'],
                    'testing': ['1.3.0'],
                    'unstable': ['1.4.0'],
                },
            )
        ]:
            spec = Spec.parse(spec)
            data = {
                'releases': dict(
                    [
                        (k, [{'version': v} for v in vs])
                        for k, vs in data.items()
                    ]
                )
            }

            self.assertEqual(res, cmd.get_best_version(data, spec))

        opt.status = Spec.UNSTABLE

        for spec, res, data in [
            (
                'foo>=1.1',
                '1.4.0',
                {
                    'stable': ['1.1.0', '1.2.0'],
                    'testing': ['1.3.0'],
                    'unstable': ['1.4.0'],
                },
            )
        ]:
            spec = Spec.parse(spec)
            data = {
                'releases': dict(
                    [
                        (k, [{'version': v} for v in vs])
                        for k, vs in data.items()
                    ]
                )
            }

            self.assertEqual(res, cmd.get_best_version(data, spec))


class Assertions(object):

    make = object()

    def assertCallArgs(self, pattern, args):
        if len(pattern) != len(args):
            self.fail('args and pattern have different lengths')
        for p, a in zip(pattern, args):
            if p is self.make:
                if not a.endswith('make'):
                    self.fail('%s is not a make in %s' % (a, args))
            else:
                if not a == p:
                    self.fail('%s is not a %s in %s' % (a, p, args))


# With mock patching a method seems tricky: looks there's no way to get to
# 'self' as the mock method is unbound.
TarArchive.unpack_orig = TarArchive.unpack
ZipArchive.unpack_orig = ZipArchive.unpack


class InstallTestCase(unittest.TestCase, Assertions):
    def setUp(self):
        self._p1 = patch('pgxnclient.network.get_file')
        self.mock_get = self._p1.start()
        self.mock_get.side_effect = fake_get_file

        self._p2 = patch('pgxnclient.commands.Popen')
        self.mock_popen = self._p2.start()
        self.mock_popen.return_value.returncode = 0

        self._p3 = patch('pgxnclient.commands.WithPgConfig.call_pg_config')
        self.mock_pgconfig = self._p3.start()
        self.mock_pgconfig.side_effect = fake_pg_config(libdir='/', bindir='/')

    def tearDown(self):
        self._p1.stop()
        self._p2.stop()
        self._p3.stop()

    def test_install_latest(self):
        from pgxnclient.cli import main

        main(['install', '--sudo', '--', 'foobar'])

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )
        self.assertCallArgs(
            ['sudo', self.make], self.mock_popen.call_args_list[1][0][0][:2]
        )

    def test_install_missing_sudo(self):
        from pgxnclient.cli import main

        self.assertRaises(InsufficientPrivileges, main, ['install', 'foobar'])

    def test_install_local(self):
        self.mock_pgconfig.side_effect = fake_pg_config(
            libdir=os.environ['HOME'], bindir='/'
        )

        from pgxnclient.cli import main

        main(['install', 'foobar'])

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[1][0][0][:1]
        )

    def test_install_url(self):
        self.mock_pgconfig.side_effect = fake_pg_config(
            libdir=os.environ['HOME'], bindir='/'
        )

        from pgxnclient.cli import main

        main(
            [
                'install',
                'https://api.pgxn.org/dist/foobar/0.42.1/foobar-0.42.1.zip',
            ]
        )

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[1][0][0][:1]
        )

    def test_install_fails(self):
        self.mock_popen.return_value.returncode = 1
        self.mock_pgconfig.side_effect = fake_pg_config(
            libdir=os.environ['HOME'], bindir='/'
        )

        from pgxnclient.cli import main

        self.assertRaises(PgxnClientException, main, ['install', 'foobar'])

        self.assertEquals(self.mock_popen.call_count, 1)

    def test_install_bad_sha1(self):
        def fakefake(url):
            return fake_get_file(
                url,
                urlmap={
                    'https://api.pgxn.org/dist/foobar/0.42.1/META.json': 'https://api.pgxn.org/dist/foobar/0.42.1/META-badsha1.json'  # noqa
                },
            )

        self.mock_get.side_effect = fakefake

        from pgxnclient.cli import main
        from pgxnclient.errors import BadChecksum

        self.assertRaises(
            BadChecksum, main, ['install', '--sudo', '--', 'foobar']
        )

    def test_install_nosudo(self):
        self.mock_pgconfig.side_effect = fake_pg_config(
            libdir=os.environ['HOME']
        )

        from pgxnclient.cli import main

        main(['install', '--nosudo', 'foobar'])

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[1][0][0][:1]
        )

    def test_install_sudo(self):
        from pgxnclient.cli import main

        main(['install', '--sudo', 'gksudo -d "hello world"', 'foobar'])

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )
        self.assertCallArgs(
            ['gksudo', '-d', 'hello world', self.make],
            self.mock_popen.call_args_list[1][0][0][:4],
        )

    @patch('pgxnclient.tar.TarArchive.unpack')
    def test_install_local_tar(self, mock_unpack):
        fn = get_test_filename('foobar-0.42.1.tar.gz')
        mock_unpack.side_effect = TarArchive(fn).unpack_orig

        from pgxnclient.cli import main

        main(['install', '--sudo', '--', fn])

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )
        self.assertCallArgs(
            ['sudo', self.make], self.mock_popen.call_args_list[1][0][0][:2]
        )
        make_cwd = self.mock_popen.call_args_list[1][1]['cwd']

        self.assertEquals(mock_unpack.call_count, 1)
        (tmpdir,) = mock_unpack.call_args[0]
        self.assertEqual(make_cwd, os.path.join(tmpdir, 'foobar-0.42.1'))

    @patch('pgxnclient.zip.ZipArchive.unpack')
    def test_install_local_zip(self, mock_unpack):
        fn = get_test_filename('foobar-0.42.1.zip')
        mock_unpack.side_effect = ZipArchive(fn).unpack_orig

        from pgxnclient.cli import main

        main(['install', '--sudo', '--', fn])

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )
        self.assertCallArgs(
            ['sudo', self.make], self.mock_popen.call_args_list[1][0][0][:2]
        )
        make_cwd = self.mock_popen.call_args_list[1][1]['cwd']

        self.assertEquals(mock_unpack.call_count, 1)
        (tmpdir,) = mock_unpack.call_args[0]
        self.assertEqual(make_cwd, os.path.join(tmpdir, 'foobar-0.42.1'))

    def test_install_url_file(self):
        fn = get_test_filename('foobar-0.42.1.zip')
        url = 'file://' + os.path.abspath(fn).replace("f", '%%%2x' % ord('f'))

        from pgxnclient.cli import main

        main(['install', '--sudo', '--', url])

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )
        self.assertCallArgs(
            ['sudo', self.make], self.mock_popen.call_args_list[1][0][0][:2]
        )

    def test_install_local_dir(self):
        self.mock_get.side_effect = lambda *args: self.fail('network invoked')

        tdir = tempfile.mkdtemp()
        try:
            from pgxnclient.zip import unpack

            dir = unpack(get_test_filename('foobar-0.42.1.zip'), tdir)

            from pgxnclient.cli import main

            main(['install', '--sudo', '--', dir])

        finally:
            shutil.rmtree(tdir)

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )
        self.assertCallArgs(dir, self.mock_popen.call_args_list[0][1]['cwd'])
        self.assertCallArgs(
            ['sudo', self.make], self.mock_popen.call_args_list[1][0][0][:2]
        )
        self.assertEquals(dir, self.mock_popen.call_args_list[1][1]['cwd'])


class CheckTestCase(unittest.TestCase, Assertions):
    def setUp(self):
        self._p1 = patch('pgxnclient.network.get_file')
        self.mock_get = self._p1.start()
        self.mock_get.side_effect = fake_get_file

        self._p2 = patch('pgxnclient.commands.Popen')
        self.mock_popen = self._p2.start()
        self.mock_popen.return_value.returncode = 0

        self._p3 = patch('pgxnclient.commands.WithPgConfig.call_pg_config')
        self.mock_pgconfig = self._p3.start()
        self.mock_pgconfig.side_effect = fake_pg_config(libdir='/', bindir='/')

    def tearDown(self):
        self._p1.stop()
        self._p2.stop()
        self._p3.stop()

    def test_check_latest(self):
        from pgxnclient.cli import main

        main(['check', 'foobar'])

        self.assertEquals(self.mock_popen.call_count, 1)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )

    def test_check_url(self):
        from pgxnclient.cli import main

        main(
            [
                'check',
                'https://api.pgxn.org/dist/foobar/0.42.1/foobar-0.42.1.zip',
            ]
        )

        self.assertEquals(self.mock_popen.call_count, 1)
        self.assertCallArgs(
            [self.make], self.mock_popen.call_args_list[0][0][0][:1]
        )

    def test_check_fails(self):
        self.mock_popen.return_value.returncode = 1

        from pgxnclient.cli import main

        self.assertRaises(PgxnClientException, main, ['check', 'foobar'])

        self.assertEquals(self.mock_popen.call_count, 1)

    def test_check_diff_moved(self):
        def create_regression_files(*args, **kwargs):
            cwd = kwargs['cwd']
            open(os.path.join(cwd, 'regression.out'), 'w').close()
            open(os.path.join(cwd, 'regression.diffs'), 'w').close()
            return Mock()

        self.mock_popen.side_effect = create_regression_files
        self.mock_popen.return_value.returncode = 1

        self.assert_(
            not os.path.exists('regression.out'),
            "Please remove temp file 'regression.out' from current dir",
        )
        self.assert_(
            not os.path.exists('regression.diffs'),
            "Please remove temp file 'regression.diffs' from current dir",
        )

        from pgxnclient.cli import main

        try:
            self.assertRaises(PgxnClientException, main, ['check', 'foobar'])
            self.assertEquals(self.mock_popen.call_count, 1)
            self.assert_(os.path.exists('regression.out'))
            self.assert_(os.path.exists('regression.diffs'))
        finally:
            ifunlink('regression.out')
            ifunlink('regression.diffs')

    def test_check_bad_sha1(self):
        def fakefake(url):
            return fake_get_file(
                url,
                urlmap={
                    'https://api.pgxn.org/dist/foobar/0.42.1/META.json': 'https://api.pgxn.org/dist/foobar/0.42.1/META-badsha1.json'  # noqa
                },
            )

        self.mock_get.side_effect = fakefake
        self.mock_popen.return_value.returncode = 1

        from pgxnclient.cli import main
        from pgxnclient.errors import BadChecksum

        self.assertRaises(BadChecksum, main, ['check', 'foobar'])

        self.assertEquals(self.mock_popen.call_count, 0)


class LoadTestCase(unittest.TestCase):
    def setUp(self):
        self._p1 = patch('pgxnclient.commands.Popen')
        self.mock_popen = self._p1.start()
        self.mock_popen.return_value.returncode = 0
        self.mock_popen.return_value.communicate.return_value = (b'', b'')

        self._p2 = patch('pgxnclient.commands.install.LoadUnload.is_extension')
        self.mock_isext = self._p2.start()
        self.mock_isext.return_value = True

        self._p3 = patch(
            'pgxnclient.commands.install.LoadUnload.get_pg_version'
        )
        self.mock_pgver = self._p3.start()
        self.mock_pgver.return_value = (9, 1, 0)

    def tearDown(self):
        self._p1.stop()
        self._p2.stop()
        self._p3.stop()

    def test_parse_version(self):
        from pgxnclient.commands.install import Load

        cmd = Load(None)
        self.assertEquals((9, 0, 3), cmd.parse_pg_version('90003'))
        self.assertEquals((9, 1, 0), cmd.parse_pg_version('90100'))
        self.assertEquals((10, 0), cmd.parse_pg_version('100000'))
        self.assertEquals((13, 2), cmd.parse_pg_version('130002'))
        self.assertEquals((13, 2), cmd.parse_pg_version('130002\n'))

    @patch('pgxnclient.network.get_file')
    def test_check_psql_options(self, mock_get):
        mock_get.side_effect = fake_get_file

        from pgxnclient.cli import main

        main(['load', '--yes', '--dbname', 'dbdb', 'foobar'])
        args = self.mock_popen.call_args[0][0]
        assert "-tAX" in args
        self.assertEqual('dbdb', args[args.index('--dbname') + 1])

        main(['load', '--yes', '-U', 'meme', 'foobar'])
        args = self.mock_popen.call_args[0][0]
        self.assertEqual('meme', args[args.index('--username') + 1])

        main(['load', '--yes', '--port', '666', 'foobar'])
        args = self.mock_popen.call_args[0][0]
        self.assertEqual('666', args[args.index('--port') + 1])

        main(['load', '--yes', '-h', 'somewhere', 'foobar'])
        args = self.mock_popen.call_args[0][0]
        self.assertEqual('somewhere', args[args.index('--host') + 1])

    @patch('pgxnclient.zip.ZipArchive.unpack')
    @patch('pgxnclient.network.get_file')
    def test_load_local_zip(self, mock_get, mock_unpack):
        mock_get.side_effect = lambda *args: self.fail('network invoked')
        mock_unpack.side_effect = ZipArchive.unpack_orig

        from pgxnclient.cli import main

        main(['load', '--yes', get_test_filename('foobar-0.42.1.zip')])

        self.assertEquals(mock_unpack.call_count, 0)
        self.assertEquals(self.mock_popen.call_count, 1)
        self.assert_('psql' in self.mock_popen.call_args[0][0][0])
        communicate = self.mock_popen.return_value.communicate
        self.assertEquals(
            communicate.call_args[0][0], b'CREATE EXTENSION foobar;'
        )

    @patch('pgxnclient.tar.TarArchive.unpack')
    @patch('pgxnclient.network.get_file')
    def test_load_local_tar(self, mock_get, mock_unpack):
        mock_get.side_effect = lambda *args: self.fail('network invoked')
        mock_unpack.side_effect = TarArchive.unpack_orig

        from pgxnclient.cli import main

        main(['load', '--yes', get_test_filename('foobar-0.42.1.tar.gz')])

        self.assertEquals(mock_unpack.call_count, 0)
        self.assertEquals(self.mock_popen.call_count, 1)
        self.assert_('psql' in self.mock_popen.call_args[0][0][0])
        communicate = self.mock_popen.return_value.communicate
        self.assertEquals(
            communicate.call_args[0][0], b'CREATE EXTENSION foobar;'
        )

    @patch('pgxnclient.network.get_file')
    def test_load_local_dir(self, mock_get):
        mock_get.side_effect = lambda *args: self.fail('network invoked')

        tdir = tempfile.mkdtemp()
        try:
            from pgxnclient.zip import unpack

            dir = unpack(get_test_filename('foobar-0.42.1.zip'), tdir)

            from pgxnclient.cli import main

            main(['load', '--yes', dir])

        finally:
            shutil.rmtree(tdir)

        self.assertEquals(self.mock_popen.call_count, 1)
        self.assert_('psql' in self.mock_popen.call_args[0][0][0])
        communicate = self.mock_popen.return_value.communicate
        self.assertEquals(
            communicate.call_args[0][0], b'CREATE EXTENSION foobar;'
        )

    @patch('pgxnclient.zip.ZipArchive.unpack')
    @patch('pgxnclient.network.get_file')
    def test_load_zip_url(self, mock_get, mock_unpack):
        mock_get.side_effect = fake_get_file
        mock_unpack.side_effect = ZipArchive.unpack_orig

        from pgxnclient.cli import main

        main(
            [
                'load',
                '--yes',
                'https://api.pgxn.org/dist/foobar/0.42.1/foobar-0.42.1.zip',
            ]
        )

        self.assertEquals(mock_unpack.call_count, 0)
        self.assertEquals(self.mock_popen.call_count, 1)
        self.assert_('psql' in self.mock_popen.call_args[0][0][0])
        communicate = self.mock_popen.return_value.communicate
        self.assertEquals(
            communicate.call_args[0][0], b'CREATE EXTENSION foobar;'
        )

    @patch('pgxnclient.tar.TarArchive.unpack')
    @patch('pgxnclient.network.get_file')
    def test_load_tar_url(self, mock_get, mock_unpack):
        mock_get.side_effect = fake_get_file
        mock_unpack.side_effect = TarArchive.unpack_orig

        from pgxnclient.cli import main

        main(['load', '--yes', 'https://example.org/foobar-0.42.1.tar.gz'])

        self.assertEquals(mock_unpack.call_count, 0)
        self.assertEquals(self.mock_popen.call_count, 1)
        self.assert_('psql' in self.mock_popen.call_args[0][0][0])
        communicate = self.mock_popen.return_value.communicate
        self.assertEquals(
            communicate.call_args[0][0], b'CREATE EXTENSION foobar;'
        )

    def test_load_extensions_order(self):
        tdir = tempfile.mkdtemp()
        try:
            from pgxnclient.zip import unpack

            dir = unpack(get_test_filename('foobar-0.42.1.zip'), tdir)
            shutil.copyfile(
                get_test_filename('META-manyext.json'),
                os.path.join(dir, 'META.json'),
            )

            from pgxnclient.cli import main

            main(['load', '--yes', dir])

        finally:
            shutil.rmtree(tdir)

        self.assertEquals(self.mock_popen.call_count, 4)
        self.assert_('psql' in self.mock_popen.call_args[0][0][0])
        communicate = self.mock_popen.return_value.communicate
        self.assertEquals(
            communicate.call_args_list[0][0][0], b'CREATE EXTENSION foo;'
        )
        self.assertEquals(
            communicate.call_args_list[1][0][0], b'CREATE EXTENSION bar;'
        )
        self.assertEquals(
            communicate.call_args_list[2][0][0], b'CREATE EXTENSION baz;'
        )
        self.assertEquals(
            communicate.call_args_list[3][0][0], b'CREATE EXTENSION qux;'
        )

    def test_unload_extensions_order(self):
        tdir = tempfile.mkdtemp()
        try:
            from pgxnclient.zip import unpack

            dir = unpack(get_test_filename('foobar-0.42.1.zip'), tdir)
            shutil.copyfile(
                get_test_filename('META-manyext.json'),
                os.path.join(dir, 'META.json'),
            )

            from pgxnclient.cli import main

            main(['unload', '--yes', dir])

        finally:
            shutil.rmtree(tdir)

        self.assertEquals(self.mock_popen.call_count, 4)
        self.assert_('psql' in self.mock_popen.call_args[0][0][0])
        communicate = self.mock_popen.return_value.communicate
        self.assertEquals(
            communicate.call_args_list[0][0][0], b'DROP EXTENSION qux;'
        )
        self.assertEquals(
            communicate.call_args_list[1][0][0], b'DROP EXTENSION baz;'
        )
        self.assertEquals(
            communicate.call_args_list[2][0][0], b'DROP EXTENSION bar;'
        )
        self.assertEquals(
            communicate.call_args_list[3][0][0], b'DROP EXTENSION foo;'
        )

    def test_load_list(self):
        tdir = tempfile.mkdtemp()
        try:
            from pgxnclient.zip import unpack

            dir = unpack(get_test_filename('foobar-0.42.1.zip'), tdir)
            shutil.copyfile(
                get_test_filename('META-manyext.json'),
                os.path.join(dir, 'META.json'),
            )

            from pgxnclient.cli import main

            main(['load', '--yes', dir, 'baz', 'foo'])

        finally:
            shutil.rmtree(tdir)

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assert_('psql' in self.mock_popen.call_args[0][0][0])
        communicate = self.mock_popen.return_value.communicate
        self.assertEquals(
            communicate.call_args_list[0][0][0], b'CREATE EXTENSION baz;'
        )
        self.assertEquals(
            communicate.call_args_list[1][0][0], b'CREATE EXTENSION foo;'
        )

    def test_unload_list(self):
        tdir = tempfile.mkdtemp()
        try:
            from pgxnclient.zip import unpack

            dir = unpack(get_test_filename('foobar-0.42.1.zip'), tdir)
            shutil.copyfile(
                get_test_filename('META-manyext.json'),
                os.path.join(dir, 'META.json'),
            )

            from pgxnclient.cli import main

            main(['unload', '--yes', dir, 'baz', 'foo'])

        finally:
            shutil.rmtree(tdir)

        self.assertEquals(self.mock_popen.call_count, 2)
        self.assert_('psql' in self.mock_popen.call_args[0][0][0])
        communicate = self.mock_popen.return_value.communicate
        self.assertEquals(
            communicate.call_args_list[0][0][0], b'DROP EXTENSION baz;'
        )
        self.assertEquals(
            communicate.call_args_list[1][0][0], b'DROP EXTENSION foo;'
        )

    def test_load_missing(self):
        tdir = tempfile.mkdtemp()
        try:
            from pgxnclient.zip import unpack

            dir = unpack(get_test_filename('foobar-0.42.1.zip'), tdir)
            shutil.copyfile(
                get_test_filename('META-manyext.json'),
                os.path.join(dir, 'META.json'),
            )

            from pgxnclient.cli import main

            self.assertRaises(
                PgxnClientException, main, ['load', '--yes', dir, 'foo', 'ach']
            )

        finally:
            shutil.rmtree(tdir)

        self.assertEquals(self.mock_popen.call_count, 0)

    def test_unload_missing(self):
        tdir = tempfile.mkdtemp()
        try:
            from pgxnclient.zip import unpack

            dir = unpack(get_test_filename('foobar-0.42.1.zip'), tdir)
            shutil.copyfile(
                get_test_filename('META-manyext.json'),
                os.path.join(dir, 'META.json'),
            )

            from pgxnclient.cli import main

            self.assertRaises(
                PgxnClientException,
                main,
                ['unload', '--yes', dir, 'foo', 'ach'],
            )

        finally:
            shutil.rmtree(tdir)

        self.assertEquals(self.mock_popen.call_count, 0)

    def test_missing_meta_dir(self):
        # issue #19
        tdir = tempfile.mkdtemp()
        try:
            from pgxnclient.zip import unpack

            dir = unpack(get_test_filename('foobar-0.42.1.zip'), tdir)
            os.unlink(os.path.join(dir, 'META.json'))

            from pgxnclient.cli import main

            self.assertRaises(PgxnClientException, main, ['load', dir])

        finally:
            shutil.rmtree(tdir)


class SearchTestCase(unittest.TestCase):
    @patch('sys.stdout')
    @patch('pgxnclient.network.get_file')
    def test_search_quoting(self, mock_get, stdout):
        stdout.encoding = 'UTF-8'
        mock_get.side_effect = fake_get_file

        from pgxnclient.cli import main

        main(['search', '--docs', 'foo bar', 'baz'])

    @patch('sys.stdout')
    @patch('pgxnclient.network.get_file')
    def test_nonascii(self, mock_get, stdout):
        mock_get.side_effect = fake_get_file

        from pgxnclient.cli import main

        stdout.encoding = 'UTF-8'
        main(['search', 'oracle'])
        out = max(get_stdout_data(stdout))
        if not isinstance(out, int):
            out = ord(out)
        assert out > 127

        stdout.reset_mock()
        stdout.encoding = None
        main(['search', 'oracle'])
        out = max(get_stdout_data(stdout))
        if not isinstance(out, int):
            out = ord(out)
        assert out < 127


class HelpTestCase(unittest.TestCase):
    @patch('sys.stdout')
    def test_libexec(self, stdout):
        stdout.encoding = 'UTF-8'

        from pgxnclient.cli import main

        main(['help', '--libexec'])
        out = get_stdout_data(stdout)
        assert out.strip()
        assert out.count(b'\n') == 1


def get_stdout_data(mock):
    # TODO: reorganize tests to be less verbose and do this kind of cruft
    # with fixtures and pytest-foo
    calls = mock.write.call_args_list or mock.buffer.write.call_args_list
    out = [a[0] for a, k in calls]
    if out and not isinstance(out[0], bytes):
        out = [x.encode("utf-8", "replace") for x in out]  # just guessing
    return b''.join(out)


if __name__ == '__main__':
    unittest.main()



import unittest

from pgxnclient import tar
from pgxnclient import zip
from pgxnclient import archive

from pgxnclient.errors import PgxnClientException
from .testutils import get_test_filename


class TestArchive(unittest.TestCase):
    def test_from_file_zip(self):
        fn = get_test_filename('foobar-0.42.1.zip')
        a = archive.from_file(fn)
        self.assert_(isinstance(a, zip.ZipArchive))
        self.assertEqual(a.filename, fn)

    def test_from_file_tar(self):
        fn = get_test_filename('foobar-0.42.1.tar.gz')
        a = archive.from_file(fn)
        self.assert_(isinstance(a, tar.TarArchive))
        self.assertEqual(a.filename, fn)

    def test_from_file_unknown(self):
        fn = get_test_filename('META-manyext.json')
        self.assertRaises(PgxnClientException, archive.from_file, fn)


class TestZipArchive(unittest.TestCase):
    def test_can_open(self):
        fn = get_test_filename('foobar-0.42.1.zip')
        a = zip.ZipArchive(fn)
        self.assert_(a.can_open())
        a.open()
        a.close()

    def test_can_open_noext(self):
        fn = get_test_filename('zip.ext')
        a = zip.ZipArchive(fn)
        self.assert_(a.can_open())
        a.open()
        a.close()

    def test_cannot_open(self):
        fn = get_test_filename('foobar-0.42.1.tar.gz')
        a = zip.ZipArchive(fn)
        self.assert_(not a.can_open())


class TestTarArchive(unittest.TestCase):
    def test_can_open(self):
        fn = get_test_filename('foobar-0.42.1.tar.gz')
        a = tar.TarArchive(fn)
        self.assert_(a.can_open())
        a.open()
        a.close()

    def test_can_open_noext(self):
        fn = get_test_filename('tar.ext')
        a = tar.TarArchive(fn)
        self.assert_(a.can_open())
        a.open()
        a.close()

    def test_cannot_open(self):
        fn = get_test_filename('foobar-0.42.1.zip')
        a = tar.TarArchive(fn)
        self.assert_(not a.can_open())



"""
pgxnclient -- network interaction
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os
from six.moves.urllib.request import build_opener
from six.moves.urllib.error import HTTPError, URLError
from six.moves.urllib.parse import urlsplit
from itertools import count
from contextlib import closing

from pgxnclient import __version__
from pgxnclient.i18n import _
from pgxnclient.errors import (
    PgxnClientException,
    NetworkError,
    ResourceNotFound,
    BadRequestError,
)

import logging

logger = logging.getLogger('pgxnclient.network')


def get_file(url):
    opener = build_opener()
    opener.addheaders = [('User-agent', 'pgxnclient/%s' % __version__)]
    logger.debug('opening url: %s', url)
    try:
        return closing(opener.open(url))
    except HTTPError as e:
        if e.code == 404:
            raise ResourceNotFound(_("resource not found: '%s'") % e.url)
        elif e.code == 400:
            raise BadRequestError(_("bad request on '%s'") % e.url)
        elif e.code == 500:
            raise NetworkError(_("server error"))
        elif e.code == 503:
            raise NetworkError(_("service unavailable"))
        else:
            raise NetworkError(
                _("unexpected response %d for '%s'") % (e.code, e.url)
            )
    except URLError as e:
        raise NetworkError(_("network error: %s") % e.reason)


def get_local_file_name(target, url):
    """Return a good name for a local file.

    If *target* is a dir, make a name out of the url. Otherwise return target
    itself. Always return an absolute path.
    """
    if os.path.isdir(target):
        basename = urlsplit(url)[2].rsplit('/', 1)[-1]
        fn = os.path.join(target, basename)
    else:
        fn = target

    return os.path.abspath(fn)


def download(f, fn, rename=True):
    """Download a file locally.

    :param f: open file to read
    :param fn: name of the file to write. If a dir, save into it.
    :param rename: if true and a file *fn* exist, rename the downloaded file
        adding a prefix ``-1``, ``-2``... before the extension.

    Return the name of the file saved.
    """
    if os.path.isdir(fn):
        fn = get_local_file_name(fn, f.url)

    if rename:
        if os.path.exists(fn):
            base, ext = os.path.splitext(fn)
            for i in count(1):
                logger.debug(_("file %s exists"), fn)
                fn = "%s-%d%s" % (base, i, ext)
                if not os.path.exists(fn):
                    break

    logger.info(_("saving %s"), fn)
    try:
        fout = open(fn, "wb")
    except Exception as e:
        raise PgxnClientException(
            _("cannot open target file: %s: %s") % (e.__class__.__name__, e)
        )
    try:
        while 1:
            data = f.read(8192)
            if not data:
                break
            fout.write(data)
    finally:
        fout.close()

    return fn



"""
pgxnclient -- informative commands implementation
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import re
import logging
import textwrap
import xml.sax.saxutils as saxutils

import six

from pgxnclient import SemVer
from pgxnclient.i18n import _, N_
from pgxnclient.utils import emit
from pgxnclient.errors import NotFound, ResourceNotFound
from pgxnclient.commands import Command, WithSpec

logger = logging.getLogger('pgxnclient.commands')


class Mirror(Command):
    name = 'mirror'
    description = N_("return information about the available mirrors")

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        subp = super(Mirror, self).customize_parser(
            parser, subparsers, **kwargs
        )

        subp.add_argument(
            'uri',
            nargs='?',
            metavar="URI",
            help=_(
                "return detailed info about this mirror."
                " If not specified return a list of mirror URIs"
            ),
        )
        subp.add_argument(
            '--detailed',
            action="store_true",
            help=_("return full details for each mirror"),
        )

        return subp

    def run(self):
        data = self.api.mirrors()
        if self.opts.uri:
            detailed = True
            data = [d for d in data if d['uri'] == self.opts.uri]
            if not data:
                raise ResourceNotFound(
                    _('mirror not found: %s') % self.opts.uri
                )
        else:
            detailed = self.opts.detailed

        for i, d in enumerate(data):
            if not detailed:
                emit(d['uri'])
            else:
                for k in u"""
                    uri frequency location bandwidth organization email
                    timezone src rsync notes
                """.split():
                    emit("%s: %s" % (k, d.get(k, '')))

                emit()


class Search(Command):
    name = 'search'
    description = N_("search in the available extensions")

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        subp = super(Search, self).customize_parser(
            parser, subparsers, **kwargs
        )

        g = subp.add_mutually_exclusive_group()
        g.add_argument(
            '--docs',
            dest='where',
            action='store_const',
            const='docs',
            default='docs',
            help=_("search in documentation [default]"),
        )
        g.add_argument(
            '--dist',
            dest='where',
            action='store_const',
            const="dists",
            help=_("search in distributions"),
        )
        g.add_argument(
            '--ext',
            dest='where',
            action='store_const',
            const='extensions',
            help=_("search in extensions"),
        )
        subp.add_argument(
            'query', metavar='TERM', nargs='+', help=_("a string to search")
        )

        return subp

    def run(self):
        data = self.api.search(self.opts.where, self.opts.query)

        for hit in data['hits']:
            emit("%s %s" % (hit['dist'], hit['version']))
            if 'excerpt' in hit:
                excerpt = self.clean_excerpt(hit['excerpt'])

                for line in textwrap.wrap(excerpt, 72):
                    emit("    " + line)
                emit()

    def clean_excerpt(self, excerpt):
        """Clean up the excerpt returned in the json result for output."""
        # replace ellipsis with three dots, as there's no chance
        # to have them printed on non-utf8 consoles.
        # Also, they suck obscenely on fixed-width output.
        excerpt = excerpt.replace('&#8230;', '...')

        # TODO: this apparently misses a few entities
        excerpt = saxutils.unescape(excerpt)
        excerpt = excerpt.replace('&quot;', '"')

        # Convert numerical entities
        excerpt = re.sub(
            r'\&\#(\d+)\;', lambda c: six.unichr(int(c.group(1))), excerpt
        )

        # Hilight found terms
        # TODO: use proper highlight with escape chars?
        excerpt = excerpt.replace('<strong></strong>', '')
        excerpt = excerpt.replace('<strong>', '*')
        excerpt = excerpt.replace('</strong>', '*')

        return excerpt


class Info(WithSpec, Command):
    name = 'info'
    description = N_("print information about a distribution")

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        subp = super(Info, self).customize_parser(parser, subparsers, **kwargs)

        g = subp.add_mutually_exclusive_group()
        g.add_argument(
            '--details',
            dest='what',
            action='store_const',
            const='details',
            default='details',
            help=_("show details about the distribution [default]"),
        )
        g.add_argument(
            '--meta',
            dest='what',
            action='store_const',
            const='meta',
            help=_("show the distribution META.json"),
        )
        g.add_argument(
            '--readme',
            dest='what',
            action='store_const',
            const='readme',
            help=_("show the distribution README"),
        )
        g.add_argument(
            '--versions',
            dest='what',
            action='store_const',
            const='versions',
            help=_("show the list of available versions"),
        )

        return subp

    def run(self):
        spec = self.get_spec()
        getattr(self, 'print_' + self.opts.what)(spec)

    def print_meta(self, spec):
        data = self._get_dist_data(spec.name)
        ver = self.get_best_version(data, spec, quiet=True)
        emit(self.api.meta(spec.name, ver, as_json=False))

    def print_readme(self, spec):
        data = self._get_dist_data(spec.name)
        ver = self.get_best_version(data, spec, quiet=True)
        emit(self.api.readme(spec.name, ver))

    def print_details(self, spec):
        data = self._get_dist_data(spec.name)
        ver = self.get_best_version(data, spec, quiet=True)
        data = self.api.meta(spec.name, ver)
        for k in u"""
            name abstract description maintainer license release_status
            version date sha1
        """.split():
            try:
                v = data[k]
            except KeyError:
                logger.warning(_("data key '%s' not found"), k)
                continue

            if isinstance(v, list):
                for vv in v:
                    emit("%s: %s" % (k, vv))
            elif isinstance(v, dict):
                for kk, vv in v.items():
                    emit("%s: %s: %s" % (k, kk, vv))
            else:
                emit("%s: %s" % (k, v))

        k = 'provides'
        for ext, dext in data[k].items():
            emit("%s: %s: %s" % (k, ext, dext['version']))

        k = 'prereqs'
        if k in data:
            for phase, rels in data[k].items():
                for rel, pkgs in rels.items():
                    for pkg, ver in pkgs.items():
                        emit("%s: %s: %s %s" % (phase, rel, pkg, ver))

    def print_versions(self, spec):
        data = self._get_dist_data(spec.name)
        name = data['name']
        vs = [
            (SemVer(d['version']), s)
            for s, ds in data['releases'].items()
            for d in ds
        ]
        vs = [(v, s) for v, s in vs if spec.accepted(v)]
        vs.sort(reverse=True)
        for v, s in vs:
            emit("%s %s %s" % (name, v, s))

    def _get_dist_data(self, name):
        try:
            return self.api.dist(name)
        except NotFound as e:
            # maybe the user was looking for an extension instead?
            try:
                ext = self.api.ext(name)
            except NotFound:
                pass
            else:
                vs = ext.get('versions', {})
                for extver, ds in vs.items():
                    for d in ds:
                        if 'dist' not in d:
                            continue
                        dist = d['dist']
                        distver = d.get('version', 'unknown')
                        logger.info(
                            _("extension %s %s found in distribution %s %s"),
                            name,
                            extver,
                            dist,
                            distver,
                        )

            raise e



#!/usr/bin/env python
"""Download an URL and save it with tne name as the urlquoted url.

The files downloaded are used by the test suite.
"""
import os
import sys
from six.moves.urllib.parse import quote
from six.moves.urllib.request import urlopen

if __name__ == '__main__':
    url = sys.argv[1]
    fn = os.path.join(os.path.dirname(__file__), quote(url, safe=''))
    f = open(fn, "wb")
    try:
        try:
            f.write(urlopen(url).read())
        finally:
            f.close()
    except Exception:
        os.unlink(fn)
        raise



"""
pgxnclient -- archives handling
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os

from pgxnclient.i18n import _
from pgxnclient.utils import load_jsons
from pgxnclient.errors import PgxnClientException


def from_spec(spec):
    """Return an `Archive` instance to handle the file requested by *spec*"""
    assert spec.is_file()
    return from_file(spec.filename)


def from_file(filename):
    """Return an `Archive` instance to handle the file *filename*"""
    from pgxnclient.zip import ZipArchive
    from pgxnclient.tar import TarArchive

    for cls in (ZipArchive, TarArchive):
        a = cls(filename)
        if a.can_open():
            return a

    raise PgxnClientException(
        _("can't open archive '%s': file type not recognized") % filename
    )


class Archive(object):
    """Base class to handle archives."""

    def __init__(self, filename):
        self.filename = filename

    def can_open(self):
        """Return `!True` if the `!filename` can be opened by the obect."""
        raise NotImplementedError

    def open(self):
        """Open the archive for usage.

        Raise PgxnClientException if the archive can't be open.
        """
        raise NotImplementedError

    def close(self):
        """Close the archive after usage."""
        raise NotImplementedError

    def list_files(self):
        """Return an iterable with the list of file names in the archive."""
        raise NotImplementedError

    def read(self, fn):
        """Return a file's data from the archive."""
        raise NotImplementedError

    def unpack(self, destdir):
        raise NotImplementedError

    def get_meta(self):
        filename = self.filename

        self.open()
        try:
            # Return the first file with the expected name
            for fn in self.list_files():
                if fn.endswith('META.json'):
                    return load_jsons(self.read(fn).decode('utf8'))
            else:
                raise PgxnClientException(
                    _("file 'META.json' not found in archive '%s'") % filename
                )
        finally:
            self.close()

    def _find_work_directory(self, destdir):
        """
        Choose the directory where to work.

        Because we are mostly a wrapper for pgxs, let's look for a makefile.
        The tar should contain a single base directory, so return the first
        dir we found containing a Makefile, alternatively just return the
        unpacked dir
        """
        for dir in os.listdir(destdir):
            for fn in ('Makefile', 'makefile', 'GNUmakefile', 'configure'):
                if os.path.exists(os.path.join(destdir, dir, fn)):
                    return os.path.join(destdir, dir)

        return destdir



"""
pgxnclient -- internationalization support
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client


def gettext(msg):
    # TODO: real l10n
    return msg


_ = gettext


def N_(msg):
    """Designate a string to be found by gettext but not to be translated."""
    return msg



#!/usr/bin/env python
"""
Simple implementation of URI-Templates
(http://bitworking.org/projects/URI-Templates/).

Some bits are inspired by or based on:

    * Joe Gregorio's example implementation
      (http://code.google.com/p/uri-templates/)

    * Addressable (http://addressable.rubyforge.org/)

Simple usage::

    >>> from pgxnclient.utils import uri

    >>> args = {'foo': 'it worked'}
    >>> uri.expand_template("http://example.com/{foo}", args)
    'http://example.com/it%20worked'

    >>> args = {'a':'foo', 'b':'bar', 'a_b':'baz'}
    >>> uri.expand_template("http://example.org/{a}{b}/{a_b}", args)
    'http://example.org/foobar/baz'

You can also use keyword arguments for a more pythonic style::

    >>> uri.expand_template("http://example.org/?q={a}", a="foo")
    'http://example.org/?q=foo'

"""

import re
import six
from six.moves.urllib.parse import quote

__all__ = ["expand_template", "TemplateSyntaxError"]


class TemplateSyntaxError(Exception):
    pass


_template_pattern = re.compile(r"{([^}]+)}")


def expand_template(template, values={}, **kwargs):
    """Expand a URI template."""
    values = values.copy()
    values.update(kwargs)
    values = percent_encode(values)
    return _template_pattern.sub(lambda m: _handle_match(m, values), template)


def _handle_match(match, values):
    op, arg, variables = parse_expansion(match.group(1))
    if op:
        try:
            return getattr(_operators, op)(variables, arg, values)
        except AttributeError:
            raise TemplateSyntaxError("Unexpected operator: %r" % op)
    else:
        assert len(variables) == 1
        key, default = list(variables.items())[0]
        return values.get(key, default)


#
# Parse an expansion
# Adapted directly from the spec (Appendix A); extra validation has been added
# to make it pass all the tests.
#

_varname_pattern = re.compile(r"^[A-Za-z0-9]\w*$")


def parse_expansion(expansion):
    """
    Parse an expansion -- the part inside {curlybraces} -- into its component
    parts. Returns a tuple of (operator, argument, variabledict).

    For example::

        >>> parse_expansion("-join|&|a,b,c=1")[0:2]
        ('join', '&')

        >>> sorted(parse_expansion("-join|&|a,b,c=1")[2].items())
        [('a', None), ('b', None), ('c', '1')]

        >>> parse_expansion("c=1")
        (None, None, {'c': '1'})

    """
    if "|" in expansion:
        (op, arg, vars_) = expansion.split("|")
        op = op[1:]
    else:
        (op, arg, vars_) = (None, None, expansion)

    vars_ = vars_.split(",")

    variables = {}
    for var in vars_:
        if "=" in var:
            (varname, vardefault) = var.split("=")
            if not vardefault:
                raise TemplateSyntaxError("Invalid variable: %r" % var)
        else:
            (varname, vardefault) = (var, None)

        if not _varname_pattern.match(varname):
            raise TemplateSyntaxError("Invalid variable: %r" % varname)
        variables[varname] = vardefault

    return (op, arg, variables)


#
# Encode an entire dictionary of values
#
def percent_encode(values):
    rv = {}
    for k, v in values.items():
        if isinstance(v, six.string_types):
            rv[k] = quote(v)
        else:
            rv[k] = [quote(s) for s in v]
    return rv


#
# Operators; see Section 3.3.
# Shoved into a class just so we have an ad hoc namespace.
#


class _operators(object):
    @staticmethod
    def opt(variables, arg, values):
        for k in variables.keys():
            v = values.get(k, None)
            if v is None or (
                not isinstance(v, six.string_types) and len(v) == 0
            ):
                continue
            else:
                return arg
        return ""

    @staticmethod
    def neg(variables, arg, values):
        if _operators.opt(variables, arg, values):
            return ""
        else:
            return arg

    @staticmethod
    def listjoin(variables, arg, values):
        k = list(variables.keys())[0]
        return arg.join(values.get(k, []))

    @staticmethod
    def join(variables, arg, values):
        return arg.join(
            [
                "%s=%s" % (k, values.get(k, default))
                for k, default in variables.items()
                if values.get(k, default) is not None
            ]
        )

    @staticmethod
    def prefix(variables, arg, values):
        k, default = list(variables.items())[0]
        v = values.get(k, default)
        if v is not None and len(v) > 0:
            return arg + v
        else:
            return ""

    @staticmethod
    def append(variables, arg, values):
        k, default = list(variables.items())[0]
        v = values.get(k, default)
        if v is not None and len(v) > 0:
            return v + arg
        else:
            return ""


#
# A bunch more tests that don't rightly fit in docstrings elsewhere
# Taken from Joe Gregorio's template_parser.py.
#
_test_pre = """
    >>> expand_template('{foo}', {})
    ''
    >>> expand_template('{foo}', {'foo': 'barney'})
    'barney'
    >>> expand_template('{foo=wilma}', {})
    'wilma'
    >>> expand_template('{foo=wilma}', {'foo': 'barney'})
    'barney'
    >>> expand_template('{-prefix|&|foo}', {})
    ''
    >>> expand_template('{-prefix|&|foo=wilma}', {})
    '&wilma'
    >>> expand_template('{-prefix||foo=wilma}', {})
    'wilma'
    >>> expand_template('{-prefix|&|foo=wilma}', {'foo': 'barney'})
    '&barney'
    >>> expand_template('{-append|/|foo}', {})
    ''
    >>> expand_template('{-append|#|foo=wilma}', {})
    'wilma#'
    >>> expand_template('{-append|&?|foo=wilma}', {'foo': 'barney'})
    'barney&?'
    >>> expand_template('{-join|/|foo}', {})
    ''
    >>> expand_template('{-join|/|foo,bar}', {})
    ''
    >>> expand_template('{-join|&|q,num}', {})
    ''
    >>> expand_template('{-join|#|foo=wilma}', {})
    'foo=wilma'
    >>> expand_template('{-join|#|foo=wilma,bar}', {})
    'foo=wilma'
    >>> expand_template('{-join|&?|foo=wilma}', {'foo': 'barney'})
    'foo=barney'
    >>> expand_template('{-listjoin|/|foo}', {})
    ''
    >>> expand_template('{-listjoin|/|foo}', {'foo': ['a', 'b']})
    'a/b'
    >>> expand_template('{-listjoin||foo}', {'foo': ['a', 'b']})
    'ab'
    >>> expand_template('{-listjoin|/|foo}', {'foo': ['a']})
    'a'
    >>> expand_template('{-listjoin|/|foo}', {'foo': []})
    ''
    >>> expand_template('{-opt|&|foo}', {})
    ''
    >>> expand_template('{-opt|&|foo}', {'foo': 'fred'})
    '&'
    >>> expand_template('{-opt|&|foo}', {'foo': []})
    ''
    >>> expand_template('{-opt|&|foo}', {'foo': ['a']})
    '&'
    >>> expand_template('{-opt|&|foo,bar}', {'foo': ['a']})
    '&'
    >>> expand_template('{-opt|&|foo,bar}', {'bar': 'a'})
    '&'
    >>> expand_template('{-opt|&|foo,bar}', {})
    ''
    >>> expand_template('{-neg|&|foo}', {})
    '&'
    >>> expand_template('{-neg|&|foo}', {'foo': 'fred'})
    ''
    >>> expand_template('{-neg|&|foo}', {'foo': []})
    '&'
    >>> expand_template('{-neg|&|foo}', {'foo': ['a']})
    ''
    >>> expand_template('{-neg|&|foo,bar}', {'bar': 'a'})
    ''
    >>> expand_template('{-neg|&|foo,bar}', {'bar': []})
    '&'
    >>> expand_template('{foo}', {'foo': ' '})
    '%20'
    >>> expand_template('{-listjoin|&|foo}', {'foo': ['&', '&', '|', '_']})
    '%26&%26&%7C&_'

    # Extra hoops to deal with unpredictable dict ordering
    >>> expand_template('{-join|#|foo=wilma,bar=barney}', {}) in \
        ('bar=barney#foo=wilma', 'foo=wilma#bar=barney')
    True

"""

_syntax_errors = """
    >>> expand_template("{fred=}")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TemplateSyntaxError: Invalid variable: 'fred='

    >>> expand_template("{f:}")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TemplateSyntaxError: Invalid variable: 'f:'

    >>> expand_template("{f<}")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TemplateSyntaxError: Invalid variable: 'f<'

    >>> expand_template("{<:}")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TemplateSyntaxError: Invalid variable: '<:'

    >>> expand_template("{<:fred,barney}")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TemplateSyntaxError: Invalid variable: '<:fred'

    >>> expand_template("{>:}")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TemplateSyntaxError: Invalid variable: '>:'

    >>> expand_template("{>:fred,barney}")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TemplateSyntaxError: Invalid variable: '>:fred'

"""

__test__ = {"test_pre": _test_pre, "syntax_errors": _syntax_errors}

if __name__ == '__main__':
    import doctest

    doctest.testmod()



"""
pgxnclient -- command line entry point
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os
import sys

from pgxnclient import find_script
from pgxnclient.i18n import _
from pgxnclient.utils import emit
from pgxnclient.errors import PgxnException, UserAbort
from pgxnclient.commands import get_option_parser, load_commands, run_command


def main(argv=None):
    """
    The program main function.

    The function is still relatively self contained: it can be called with
    arguments and raises whatever exception, so it's the best entry point
    for whole system testing.
    """
    if argv is None:
        argv = sys.argv[1:]

    load_commands()
    parser = get_option_parser()
    opt = parser.parse_args(argv)
    if hasattr(opt, 'cmd'):
        run_command(opt, parser)
    else:
        parser.print_help()


def script():
    """
    Execute the program as a script.

    Set up logging, invoke main() using the user-provided arguments and handle
    any exception raised.
    """
    # Setup logging
    import logging

    logging.basicConfig(
        format="%(levelname)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
    )
    logger = logging.getLogger()

    # Dispatch to the command according to the script name
    script = sys.argv[0]
    args = sys.argv[1:]
    if os.path.basename(script).startswith('pgxn-'):
        args.insert(0, os.path.basename(script)[5:])
        # for help print
        sys.argv[0] = os.path.join(os.path.dirname(script), 'pgxn')

    # Execute the script
    try:
        main(args)

    # Different ways to fail
    except UserAbort as e:
        # The user replied "no" to some question
        logger.info("%s", e)
        sys.exit(1)

    except PgxnException as e:
        # An regular error from the program
        logger.error("%s", e)
        sys.exit(1)

    except SystemExit as e:
        # Usually the arg parser bailing out.
        if isinstance(getattr(e, 'code', None), int):
            sys.exit(e.code)
        else:
            sys.exit(1)

    except Exception:
        logger.exception(_("unexpected error"))
        sys.exit(1)

    except BaseException:
        # ctrl-c
        sys.exit(1)


def command_dispatch(argv=None):
    """
    Entry point for a script to dispatch commands to external scripts.

    Upon invocation of a command ``pgxn cmd --arg``, locate pgxn-cmd and
    execute it with --arg arguments.
    """
    if argv is None:
        argv = sys.argv[1:]

    # Assume the first arg after the option is the command to run
    for icmd, cmd in enumerate(argv):
        if not cmd.startswith('-'):
            argv = [_get_exec(cmd)] + argv[:icmd] + argv[icmd + 1 :]
            break
    else:
        # No command specified: dispatch to the pgxnclient script
        # to print basic help, main command etc.
        argv = [
            os.path.join(os.path.dirname(sys.argv[0]), 'pgxnclient')
        ] + argv

    if not os.access(argv[0], os.X_OK):
        # This is our friend setuptools' job: the script have lost the
        # executable flag. We assume the script is a Python one and run it
        # through the current executable.
        argv.insert(0, sys.executable)

    os.execv(argv[0], argv)


def _get_exec(cmd):
    fn = find_script('pgxn-' + cmd)
    if not fn:
        emit(
            "pgxn: unknown command: '%s'. See 'pgxn --help'" % cmd,
            file=sys.stderr,
        )
        sys.exit(2)

    return fn


if __name__ == '__main__':
    script()



"""
pgxnclient -- client API stub
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

from six.moves.urllib.parse import urlencode

from pgxnclient import network
from pgxnclient.utils import load_json
from pgxnclient.errors import NetworkError, NotFound, ResourceNotFound
from pgxnclient.utils.uri import expand_template


class Api(object):
    def __init__(self, mirror):
        self.mirror = mirror

    def dist(self, dist, version=''):
        try:
            with self.call(
                version and 'meta' or 'dist',
                {'dist': dist, 'version': version},
            ) as f:
                return load_json(f)
        except ResourceNotFound:
            raise NotFound("distribution '%s' not found" % dist)

    def ext(self, ext):
        try:
            with self.call('extension', {'extension': ext}) as f:
                return load_json(f)
        except ResourceNotFound:
            raise NotFound("extension '%s' not found" % ext)

    def meta(self, dist, version, as_json=True):
        with self.call('meta', {'dist': dist, 'version': version}) as f:
            if as_json:
                return load_json(f)
            else:
                return f.read().decode('utf-8')

    def readme(self, dist, version):
        with self.call('readme', {'dist': dist, 'version': version}) as f:
            return f.read()

    def download(self, dist, version):
        dist = dist.lower()
        version = version.lower()
        return self.call('download', {'dist': dist, 'version': version})

    def mirrors(self):
        with self.call('mirrors') as f:
            return load_json(f)

    def search(self, where, query):
        """Search into PGXN.

        :param where: where to search. The server currently supports "docs",
            "dists", "extensions"
        :param query: list of strings to search
        """
        # convert the query list into a string
        q = ' '.join([' ' in s and ('"%s"' % s) or s for s in query])

        with self.call('search', {'in': where}, query={'q': q}) as f:
            return load_json(f)

    def stats(self, arg):
        with self.call('stats', {'stats': arg}) as f:
            return load_json(f)

    def user(self, username):
        with self.call('user', {'user': username}) as f:
            return load_json(f)

    def call(self, meth, args=None, query=None):
        url = self.get_url(meth, args, query)
        try:
            return network.get_file(url)
        except ResourceNotFound:
            # check if it is one of the broken URLs as reported in
            # https://groups.google.com/group/pgxn-users/browse_thread/thread/e41fbc202680c92c
            version = args and args.get('version')
            if not (version and version.trail):
                raise

            args = args.copy()
            args['version'] = str(version).replace('-', '', 1)
            url = self.get_url(meth, args, query)
            return network.get_file(url)

    def get_url(self, meth, args=None, query=None):
        tmpl = self.get_template(meth)
        url = expand_template(tmpl, args or {})
        url = self.mirror.rstrip('/') + url
        if query is not None:
            url = url + '?' + urlencode(query)

        return url

    def get_template(self, meth):
        return self.get_index()[meth]

    _api_index = None

    def get_index(self):
        if self._api_index is None:
            url = self.mirror.rstrip('/') + '/index.json'
            try:
                with network.get_file(url) as f:
                    self._api_index = load_json(f)
            except ResourceNotFound:
                raise NetworkError("API index not found at '%s'" % url)

        return self._api_index



"""
pgxnclient -- specification object
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client


import os
import re
from six.moves.urllib.parse import unquote_plus
import operator as _op

from pgxnclient.i18n import _
from pgxnclient.errors import BadSpecError, ResourceNotFound

from pgxnclient.utils.semver import SemVer
from pgxnclient.utils.strings import Term


class Spec(object):
    """A name together with a range of versions."""

    # Available release statuses.
    # Order matters.
    UNSTABLE = 0
    TESTING = 1
    STABLE = 2

    STATUS = {'unstable': UNSTABLE, 'testing': TESTING, 'stable': STABLE}

    def __init__(
        self,
        name=None,
        op=None,
        ver=None,
        dirname=None,
        filename=None,
        url=None,
    ):
        self.name = name and name.lower()
        self.op = op
        self.ver = ver

        # point to local files or specific resources
        self.dirname = dirname
        self.filename = filename
        self.url = url

    def is_name(self):
        return self.name is not None

    def is_dir(self):
        return self.dirname is not None

    def is_file(self):
        return self.filename is not None

    def is_url(self):
        return self.url is not None

    def is_local(self):
        return self.is_dir() or self.is_file()

    def __str__(self):
        name = self.name or self.filename or self.dirname or self.url or "???"
        if self.op is None:
            return name
        else:
            return "%s%s%s" % (name, self.op, self.ver)

    @classmethod
    def parse(self, spec):
        """Parse a spec string into a populated Spec instance.

        Raise BadSpecError if couldn't parse.
        """
        # check if it's a network resource
        if spec.startswith('http://') or spec.startswith('https://'):
            return Spec(url=spec)

        # check if it's a local resource
        if spec.startswith('file://'):
            try_file = unquote_plus(spec[len('file://') :])
        elif os.sep in spec:
            try_file = spec
        else:
            try_file = None

        if try_file:
            # This is a local thing, let's see what
            if os.path.isdir(try_file):
                return Spec(dirname=try_file)
            elif os.path.exists(try_file):
                return Spec(filename=try_file)
            else:
                raise ResourceNotFound(_("cannot find '%s'") % try_file)

        # so we think it's a PGXN spec

        # split operator/version and name
        m = re.match(r'(.+?)(?:(==|=|>=|>|<=|<)(.*))?$', spec)
        if m is None:
            raise BadSpecError(
                _("bad format for version specification: '%s'"), spec
            )

        name = Term(m.group(1))
        op = m.group(2)
        if op == '=':
            op = '=='

        if op is not None:
            ver = SemVer.clean(m.group(3))
        else:
            ver = None

        return Spec(name, op, ver)

    def accepted(
        self,
        version,
        _map={
            '==': _op.eq,
            '<=': _op.le,
            '<': _op.lt,
            '>=': _op.ge,
            '>': _op.gt,
        },
    ):
        """Return True if the given version is accepted in the spec."""
        if self.op is None:
            return True
        return _map[self.op](version, self.ver)



"""
pgxnclient -- misc utilities package
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client


from __future__ import print_function

__all__ = ['emit', 'load_json', 'load_jsons', 'sha1', 'find_executable']


import os
import sys
import json
from collections import OrderedDict

# Import the sha1 object without warnings
from hashlib import sha1

import six


def emit(s=b'', file=None):
    """
    Print a string

    Easy yes? No. Because if the string is unicode and we are piping stdout
    into something we end up with sys.stdout.encoding = None and Python trying
    to use ascii to encode, barfing on the first accented letter (hello Jan).
    """
    if file is None:
        file = sys.stdout

    enc = file.encoding or 'ascii'

    if isinstance(s, six.text_type):
        s = s.encode(enc, 'replace')

    # OTOH, printing bytes on Py3 to stdout/stderr will barf as well...
    # It's facepalms all the way down.
    if hasattr(file, 'buffer'):
        file = file.buffer

    file.write(s)
    file.write(b'\n')


def load_json(f):
    data = f.read()
    if not isinstance(data, six.text_type):
        data = data.decode('utf-8')
    return load_jsons(data)


def load_jsons(data):
    return json.loads(data, object_pairs_hook=OrderedDict)


def find_executable(name):
    """
    Find executable by ``name`` by inspecting PATH environment variable, return
    ``None`` if nothing found.
    """
    for dir in os.environ.get('PATH', '').split(os.pathsep):
        if not dir:
            continue
        fn = os.path.abspath(os.path.join(dir, name))
        if os.path.exists(fn):
            return os.path.abspath(fn)



"""
pgxnclient -- unit test utilities
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os


def ifunlink(fn):
    """Delete a file if exists."""
    if os.path.exists(fn):
        os.unlink(fn)


def get_test_filename(*parts):
    """Return the complete file name for a testing file."""
    return os.path.join(os.path.dirname(__file__), 'testdata', *parts)



"""
pgxnclient -- help commands implementation
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os

from pgxnclient import get_scripts_dirs, get_public_scripts_dir
from pgxnclient.i18n import _, N_
from pgxnclient.utils import emit
from pgxnclient.commands import Command


class Help(Command):
    name = 'help'
    description = N_("display help and other program information")

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        subp = super(Help, self).customize_parser(parser, subparsers, **kwargs)

        g = subp.add_mutually_exclusive_group()
        g.add_argument(
            '--all',
            action="store_true",
            help=_("list all the available commands"),
        )
        g.add_argument(
            '--libexec',
            action="store_true",
            help=_("print the location of the scripts directory"),
        )
        g.add_argument(
            'command',
            metavar='CMD',
            nargs='?',
            help=_("the command to get help about"),
        )

        # To print the basic help
        self._parser = parser

        return subp

    def run(self):
        if self.opts.command:
            from pgxnclient.cli import main

            main([self.opts.command, '--help'])
        elif self.opts.all:
            self.print_all_commands()
        elif self.opts.libexec:
            self.print_libexec()
        else:
            self._parser.print_help()

    def print_all_commands(self):
        cmds = self.find_all_commands()
        title = _("Available PGXN Client commands")
        emit(title)
        emit("-" * len(title))

        for cmd in cmds:
            emit("  " + cmd)

    def find_all_commands(self):
        rv = []
        path = os.environ.get('PATH', '').split(os.pathsep)
        path[0:0] = get_scripts_dirs()
        for p in path:
            try:
                files = os.listdir(p)
            except OSError:
                # Dir missing, or not readable
                continue

            for fn in files:
                if fn.startswith('pgxn-'):
                    rv.append(fn[5:])

        rv.sort()
        return rv

    def print_libexec(self):
        emit(get_public_scripts_dir())



import unittest

from pgxnclient import Spec


class SpecTestCase(unittest.TestCase):
    def test_str(self):
        self.assertEqual(str(Spec('foo')), 'foo')
        self.assertEqual(str(Spec('foo>2.0')), 'foo>2.0')
        self.assertEqual(str(Spec('foo>2.0')), 'foo>2.0')
        self.assertEqual(str(Spec(dirname='/foo')), '/foo')
        self.assertEqual(str(Spec(dirname='/foo/foo.zip')), '/foo/foo.zip')


if __name__ == '__main__':
    unittest.main()



"""
SemVer -- (not quite) semantic version specification

http://semver.org/

IMPORTANT: don't trust this implementation. And don't trust SemVer AT ALL.
We have a bloody mess because the specification changed after being published
and after several extension had been uploaded with a version number that
suddenly had become no more valid.

https://github.com/mojombo/semver.org/issues/49

My plea for forking the spec and keep our schema has been ignored. So this
module only tries to make sure people can use PGXN, not to be conform to an
half-aborted specification.  End of rant.

This implementation is conform to the SemVer 0.3.0 implementation by David
Wheeler (http://pgxn.org/dist/semver/0.3.0/) and passes all its unit test.

Note that it is slightly non conform to the original specification, as the
trailing part should be compared in ascii order while our comparison is not
case sensitive. David has already stated that the meaning is independent on
the case (http://blog.pgxn.org/post/4948135198/case-insensitivity) and I'm
fine with that: the important thing is that the client and the server
understand each other.

"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import re
import operator

from pgxnclient.i18n import _


class SemVer(str):
    """A string representing a semantic version number.

    Non valid version numbers raise ValueError.
    """

    def __new__(cls, value):
        self = str.__new__(cls, value)
        self.tuple = SemVer.parse(value)
        return self

    @property
    def major(self):
        return self.tuple[0]

    @property
    def minor(self):
        return self.tuple[1]

    @property
    def patch(self):
        return self.tuple[2]

    @property
    def trail(self):
        return self.tuple[3]

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, str(self))

    def __eq__(self, other):
        if isinstance(other, SemVer):
            return (
                self.tuple[:3] == other.tuple[:3]
                and self.tuple[3].lower() == other.tuple[3].lower()
            )
        elif isinstance(other, str):
            return self == SemVer(other)
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.tuple[:3] + (self.tuple[3].lower(),))

    def _ltgt(self, other, op):
        if isinstance(other, SemVer):
            t1 = self.tuple[:3]
            t2 = other.tuple[:3]
            if t1 != t2:
                return op(t1, t2)

            s1 = self.tuple[3].lower()
            s2 = other.tuple[3].lower()
            if s1 == s2:
                return False
            if s1 and s2:
                return op(s1, s2)
            return op(bool(s2), bool(s1))

        elif isinstance(other, str):
            return op(self, SemVer(other))
        else:
            return NotImplemented

    def __lt__(self, other, op=operator.lt):
        return self._ltgt(other, operator.lt)

    def __gt__(self, other):
        return self._ltgt(other, operator.gt)

    def __ge__(self, other):
        return not self < other

    def __le__(self, other):
        return not other < self

    @classmethod
    def parse(self, s):
        """
        Split a valid version number in components (major, minor, patch, trail).
        """
        m = re_semver.match(s)
        if m is None:
            raise ValueError(_("bad version number: '%s'") % s)

        maj, min, patch, trail = m.groups()
        if not patch:
            patch = 0
        if not trail:
            trail = ''
        return (int(maj), int(min), int(patch), trail)

    @classmethod
    def clean(self, s):
        """
        Convert an invalid but still recognizable version number into a SemVer.
        """
        m = re_clean.match(s.strip())
        if m is None:
            raise ValueError(_("bad version number: '%s' - can't clean") % s)

        maj, min, patch, trail = m.groups()
        maj = maj and int(maj) or 0
        min = min and int(min) or 0
        patch = patch and int(patch) or 0
        trail = trail and '-' + trail.strip() or ''
        return "%d.%d.%d%s" % (maj, min, patch, trail)


re_semver = re.compile(
    r"""
    ^
        (0|[1-9][0-9]*)
    \.  (0|[1-9][0-9]*)
    \.  (0|[1-9][0-9]*)
    (?:
        -?                       # should be mandatory, but see rant above
        ([a-z][a-z0-9-]*)
    )?
    $
    """,
    re.IGNORECASE | re.VERBOSE,
)

re_clean = re.compile(
    r"""
    ^
        ([0-9]+)?
    \.? ([0-9]+)?
    \.? ([0-9]+)?
    (?:
        -?  \s*
        ([a-z][a-z0-9-]*)
    )?
    $
    """,
    re.IGNORECASE | re.VERBOSE,
)



"""
pgxnclient -- tar file utilities
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os
import tarfile

from pgxnclient.i18n import _
from pgxnclient.errors import PgxnClientException
from pgxnclient.archive import Archive

import logging

logger = logging.getLogger('pgxnclient.tar')


class TarArchive(Archive):
    """Handle .tar archives"""

    _file = None

    def can_open(self):
        return tarfile.is_tarfile(self.filename)

    def open(self):
        assert not self._file, "archive already open"
        try:
            self._file = tarfile.open(self.filename, 'r')
        except Exception as e:
            raise PgxnClientException(
                _("cannot open archive '%s': %s") % (self.filename, e)
            )

    def close(self):
        if self._file is not None:
            self._file.close()
            self._file = None

    def list_files(self):
        assert self._file, "archive not open"
        return self._file.getnames()

    def read(self, fn):
        assert self._file, "archive not open"
        return self._file.extractfile(fn).read()

    def unpack(self, destdir):
        tarname = self.filename
        logger.info(_("unpacking: %s"), tarname)
        destdir = os.path.abspath(destdir)
        self.open()
        try:
            for fn in self.list_files():
                fname = os.path.abspath(os.path.join(destdir, fn))
                if not fname.startswith(destdir):
                    raise PgxnClientException(
                        _("archive file '%s' trying to escape!") % fname
                    )

            self._file.extractall(path=destdir)
        finally:
            self.close()

        return self._find_work_directory(destdir)


def unpack(filename, destdir):
    return TarArchive(filename).unpack(destdir)



"""
Strings -- implementation of a few specific string subclasses.
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client


import re

from pgxnclient.i18n import _
from argparse import ArgumentTypeError


class CIStr(str):
    """
    A case preserving string with non case-sensitive comparison.
    """

    def __eq__(self, other):
        if isinstance(other, CIStr):
            return self.lower() == other.lower()
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        if isinstance(other, CIStr):
            return self.lower() < other.lower()
        else:
            return NotImplemented

    def __gt__(self, other):
        return other < self

    def __le__(self, other):
        return not other < self

    def __ge__(self, other):
        return not self < other


class Label(CIStr):
    """A string following the rules in RFC 1034.

    Labels can then be used as host names in domains.

    https://tools.ietf.org/html/rfc1034

    "The labels must follow the rules for ARPANET host names. They must
    start with a letter, end with a letter or digit, and have as interior
    characters only letters, digits, and hyphen. There are also some
    restrictions on the length. Labels must be 63 characters or less."

    """

    def __new__(cls, value):
        if not Label._re_chk.match(value):
            raise ValueError(_("bad label: '%s'") % value)
        return CIStr.__new__(cls, value)

    _re_chk = re.compile(r'^[a-z]([-a-z0-9]{0,61}[a-z0-9])?$', re.IGNORECASE)


class Term(CIStr):
    r"""
    A Term is a subtype of String that must be at least two characters long
    contain no slash (/), backslash (\), control, or space characters.

    See https://pgxn.org/spec#Term
    """

    def __new__(cls, value):
        if not Term._re_chk.match(value) or min(map(ord, value)) < 32:
            raise ValueError(_("not a valid term term: '%s'") % value)
        return CIStr.__new__(cls, value)

    _re_chk = re.compile(r'^[^\s/\\]{2,}$')


class Identifier(CIStr):
    """
    A string modeling a PostgreSQL identifier.
    """

    def __new__(cls, value):
        if not value:
            raise ValueError("PostgreSQL identifiers cannot be blank")
        if not Identifier._re_chk.match(value):
            value = '"%s"' % value.replace('"', '""')
        # TODO: identifier are actually case sensitive if quoted
        return CIStr.__new__(cls, value)

    _re_chk = re.compile(r'^[a-z_][a-z0-9_\$]*$', re.IGNORECASE)

    @classmethod
    def parse_arg(self, s):
        """
        Parse an Identifier from a command line argument.
        """
        try:
            return Identifier(s)
        except ValueError as e:
            # shouldn't happen anymore as we quote invalid identifiers
            raise ArgumentTypeError(e)



#!/usr/bin/env python
"""
pgxnclient -- setup script
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client


import os
import sys
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py

here = os.path.dirname(__file__)

# Grab the version without importing the module
# or we will get import errors on install if prerequisites are still missing
with open(os.path.join(here, 'pgxnclient', '__init__.py')) as f:
    for line in f:
        if line.startswith('__version__ ='):
            version = line.split("'")[1]
            break
    else:
        raise ValueError('cannot find __version__ in the pgxnclient module')

# Read the description from the readme
with open(os.path.join(here, 'README.rst')) as f:
    long_description = f.read()

# External dependencies, depending on the Python version
requires = ['six']
setup_requires = ['pytest-runner']
tests_require = ['mock', 'pytest']

if sys.version_info < (2, 7):
    raise ValueError("PGXN client requires at least Python 2.7")
if (3,) < sys.version_info < (3, 4):
    raise ValueError("PGXN client requires at least Python 3.4")


classifiers = """
Development Status :: 5 - Production/Stable
Environment :: Console
Intended Audience :: Developers
Intended Audience :: System Administrators
License :: OSI Approved :: BSD License
Operating System :: POSIX
Programming Language :: Python :: 2
Programming Language :: Python :: 3
Topic :: Database
"""


class CustomBuildPy(build_py):
    def run(self):
        build_py.run(self)
        self.fix_libexec_hashbangs()

    def fix_libexec_hashbangs(self):
        """Replace the hashbangs of the scripts in libexec."""
        for package, src_dir, build_dir, filenames in self.data_files:
            if package != 'pgxnclient':
                continue
            for filename in filenames:
                if not filename.startswith('libexec/'):
                    continue
                self.fix_script_hashbang(os.path.join(build_dir, filename))

    def fix_script_hashbang(self, filename):
        """Replace the hashbangs of a script in libexec."""
        if not os.path.exists(filename):
            return
        with open(filename) as f:
            data = f.read()
        if not data.startswith('#!'):
            return
        lines = data.splitlines()
        if 'python' not in lines[0]:
            return

        lines[0] = '#!%s' % sys.executable
        with open(filename, 'w') as f:
            for line in lines:
                f.write(line)
                f.write('\n')


setup(
    name='pgxnclient',
    description=(
        'A command line tool to interact with the PostgreSQL Extension Network.'
    ),
    long_description=long_description,
    author='Daniele Varrazzo',
    author_email='daniele.varrazzo@gmail.com',
    url='https://pgxn.github.io/pgxnclient/',
    project_urls={
        'Source': 'https://github.com/pgxn/pgxnclient',
        'Documentation': 'https://pgxn.github.io/pgxnclient/',
        'Bug Tracker': 'https://github.com/pgxn/pgxnclient/issues',
        'Discussion group': 'https://groups.google.com/group/pgxn-users/',
        'Funding': 'https://github.com/sponsors/dvarrazzo',
    },
    license='BSD',
    # NOTE: keep consistent with docs/install.txt
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
    packages=find_packages(exclude=["tests"]),
    package_data={'pgxnclient': ['libexec/*']},
    entry_points={
        'console_scripts': [
            'pgxn = pgxnclient.cli:command_dispatch',
            'pgxnclient = pgxnclient.cli:script',
        ]
    },
    classifiers=[x for x in classifiers.split('\n') if x],
    zip_safe=False,  # because we dynamically look for commands
    install_requires=requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    version=version,
    cmdclass={'build_py': CustomBuildPy},
    extras_require={'dev': ['pytest', 'mock', 'black']},
)



"""
pgxnclient -- commands package

This module contains base classes and functions to implement and deal with
commands. Concrete commands implementations are available in other package
modules.
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os
import sys
import shlex
import logging
import argparse
from subprocess import Popen, PIPE

import six

from pgxnclient.utils import load_json, find_executable

from pgxnclient import __version__
from pgxnclient import network
from pgxnclient import Spec, SemVer
from pgxnclient import archive
from pgxnclient.api import Api
from pgxnclient.i18n import _, gettext
from pgxnclient.errors import (
    BadSpecError,
    NotFound,
    PgxnClientException,
    ProcessError,
    ResourceNotFound,
    UserAbort,
)
from pgxnclient.utils.temp import temp_dir

logger = logging.getLogger('pgxnclient.commands')


def get_option_parser():
    """
    Return an option parser populated with the available commands.

    The parser is populated with all the options defined by the implemented
    commands.  Only commands defining a ``name`` attribute are added.
    The function relies on the `Command` subclasses being already
    created: call `load_commands()` before calling this function.
    """
    parser = argparse.ArgumentParser(
        # usage = _("%(prog)s [global options] COMMAND [command options]"),
        description=_(
            "Interact with the PostgreSQL Extension Network (PGXN)."
        ),
        add_help=False,
    )
    parser.add_argument(
        "--version",
        action='version',
        version="%%(prog)s %s" % __version__,
        help=_("print the version number and exit"),
    )

    # Drop the conflicting -h argument
    parser.add_argument(
        "--help",
        action='help',
        default=argparse.SUPPRESS,
        help=_('show this help message and exit'),
    )

    subparsers = parser.add_subparsers(
        title=_("available commands"),
        metavar='COMMAND',
        help=_(
            "the command to execute."
            " The complete list is available using `pgxn help --all`."
            " Builtin commands are:"
        ),
    )

    clss = [cls for cls in CommandType.subclasses if cls.name]
    clss.sort(key=lambda c: c.name)
    for cls in clss:
        cls.customize_parser(parser, subparsers)

    return parser


def load_commands():
    """
    Load all the commands known by the program.

    Currently commands are read from modules into the `pgxnclient.commands`
    package.

    Importing the package causes the `Command` classes to be created: they
    register themselves thanks to the `CommandType` metaclass.
    """
    pkgdir = os.path.dirname(__file__)
    for fn in os.listdir(pkgdir):
        if fn.startswith('_'):
            continue
        modname = __name__ + '.' + os.path.splitext(fn)[0]

        # skip already imported modules
        if modname in sys.modules:
            continue

        try:
            __import__(modname)
        except Exception as e:
            logger.warning(
                _("error importing commands module %s: %s - %s"),
                modname,
                e.__class__.__name__,
                e,
            )


def run_command(opts, parser):
    """Run the command specified by options parsed on the command line."""
    # setup the logging
    logging.getLogger().setLevel(
        opts.verbose and logging.DEBUG or logging.INFO
    )
    return opts.cmd(opts, parser=parser).run()


class CommandType(type):
    """
    Metaclass for the Command class.

    This metaclass allows self-registration of the commands: any Command
    subclass is automatically added to the `subclasses` list.
    """

    subclasses = []

    def __new__(cls, name, bases, dct):
        rv = type.__new__(cls, name, bases, dct)
        CommandType.subclasses.append(rv)
        return rv

    def __init__(cls, name, bases, dct):
        super(CommandType, cls).__init__(name, bases, dct)


class Command(six.with_metaclass(CommandType, object)):
    """
    Base class to implement client commands.

    Provide the argument parsing framework and API dispatch.

    Commands should subclass this class and possibly other mixin classes, set
    a value for the `name` and `description` arguments and implement the
    `run()` method. If command line parser customization is required,
    `customize_parser()` should be extended.
    """

    name = None
    description = None

    def __init__(self, opts, parser=None):
        """Initialize a new Command.

        The parser will be specified if the class has been initialized
        by that parser itself, so run() can expect it being not None.
        """
        self.opts = opts
        self.parser = parser
        self._api = None

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        """Customise the option parser.

        :param parser: the option parser to be customized
        :param subparsers: the action object where to register a command subparser
        :return: the new subparser created

        Subclasses should extend this method in order to add new options or a
        subparser implementing a new command. Be careful in calling the
        superclass' `customize_parser()` via `super()` in order to call all
        the mixins methods. Also note that the method must be a classmethod.
        """
        return self.__make_subparser(parser, subparsers, **kwargs)

    def run(self):
        """The actions to take when the command is invoked."""
        raise NotImplementedError

    @classmethod
    def __make_subparser(
        self, parser, subparsers, description=None, epilog=None
    ):
        """Create a new subparser with help populated."""
        subp = subparsers.add_parser(
            self.name,
            help=gettext(self.description),
            description=description or gettext(self.description),
            add_help=False,
            epilog=epilog,
        )
        subp.set_defaults(cmd=self)

        # Drop the conflicting -h argument
        subp.add_argument(
            "--help",
            action='help',
            default=argparse.SUPPRESS,
            help=_('show this help message and exit'),
        )

        glb = subp.add_argument_group(_("global options"))
        glb.add_argument(
            "--mirror",
            metavar="URL",
            default='https://api.pgxn.org/',
            help=_("the mirror to interact with [default: %(default)s]"),
        )
        glb.add_argument(
            "--verbose", action='store_true', help=_("print more information")
        )
        glb.add_argument(
            "--yes",
            action='store_true',
            help=_("assume affirmative answer to all questions"),
        )

        return subp

    @property
    def api(self):
        """Return an `Api` instance to communicate with PGXN.

        Use the value provided with ``--mirror`` to decide where to connect.
        """
        if self._api is None:
            self._api = Api(mirror=self.opts.mirror)

        return self._api

    def confirm(self, prompt):
        """Prompt an user confirmation.

        Raise `UserAbort` if the user replies "no".

        The method is no-op if the ``--yes`` option is specified.
        """
        if self.opts.yes:
            return True

        while 1:
            ans = six.input(_("%s [y/N] ") % prompt)
            if _('no').startswith(ans.lower()):
                raise UserAbort(_("operation interrupted on user request"))
            elif _('yes').startswith(ans.lower()):
                return True
            else:
                prompt = _("Please answer yes or no")

    def popen(self, cmd, *args, **kwargs):
        """
        Excecute subprocess.Popen.

        Commands should use this method instead of importing subprocess.Popen:
        this allows replacement with a mock in the test suite.
        """
        logger.debug("running command: %s", cmd)
        try:
            return Popen(cmd, *args, **kwargs)
        except OSError as e:
            if not isinstance(cmd, six.string_types):
                cmd = ' '.join(cmd)
            msg = _("%s running command: %s") % (e, cmd)
            raise ProcessError(msg)


class WithSpec(Command):
    """Mixin to implement commands taking a package specification.

    This class adds a positional argument SPEC to the parser and related
    options.
    """

    @classmethod
    def customize_parser(
        self, parser, subparsers, with_status=True, epilog=None, **kwargs
    ):
        """
        Add the SPEC related options to the parser.

        If *with_status* is true, options ``--stable``, ``--testing``,
        ``--unstable`` are also handled.
        """
        epilog = (
            _(
                """
SPEC can either specify just a name or contain required versions
indications, for instance 'pkgname=1.0', or 'pkgname>=2.1'.
"""
            )
            + (epilog or "")
        )

        subp = super(WithSpec, self).customize_parser(
            parser, subparsers, epilog=epilog, **kwargs
        )

        subp.add_argument(
            'spec',
            metavar='SPEC',
            help=_("name and optional version of the package"),
        )

        if with_status:
            g = subp.add_mutually_exclusive_group(required=False)
            g.add_argument(
                '--stable',
                dest='status',
                action='store_const',
                const=Spec.STABLE,
                default=Spec.STABLE,
                help=_("only accept stable distributions [default]"),
            )
            g.add_argument(
                '--testing',
                dest='status',
                action='store_const',
                const=Spec.TESTING,
                help=_("accept testing distributions too"),
            )
            g.add_argument(
                '--unstable',
                dest='status',
                action='store_const',
                const=Spec.UNSTABLE,
                help=_("accept unstable distributions too"),
            )

        return subp

    def get_spec(self, _can_be_local=False, _can_be_url=False):
        """
        Return the package specification requested.

        Return a `Spec` instance.
        """
        spec = self.opts.spec

        try:
            spec = Spec.parse(spec)
        except (ValueError, BadSpecError) as e:
            self.parser.error(_("cannot parse package '%s': %s") % (spec, e))

        if not _can_be_local and spec.is_local():
            raise PgxnClientException(
                _("you cannot use a local resource with this command")
            )

        if not _can_be_url and spec.is_url():
            raise PgxnClientException(
                _("you cannot use an url with this command")
            )

        return spec

    def get_best_version(self, data, spec, quiet=False):
        """
        Return the best version an user may want for a distribution.

        Return a `SemVer` instance.

        Raise `ResourceNotFound` if no version is found with the provided
        specification and options.
        """
        drels = data['releases']

        # Get the maximum version for each release status satisfying the spec
        vers = [None] * len(Spec.STATUS)
        for n, d in drels.items():
            vs = list(filter(spec.accepted, [SemVer(r['version']) for r in d]))
            if vs:
                vers[Spec.STATUS[n]] = max(vs)

        return self._get_best_version(vers, spec, quiet)

    def get_best_version_from_ext(self, data, spec):
        """
        Return the best distribution version from an extension's data
        """
        # Get the maximum version for each release status satisfying the spec
        vers = [[] for i in range(len(Spec.STATUS))]
        vmap = {}  # ext_version -> (dist_name, dist_version)
        for ev, dists in data.get('versions', {}).items():
            ev = SemVer(ev)
            if not spec.accepted(ev):
                continue
            for dist in dists:
                dv = SemVer(dist['version'])
                ds = dist.get('status', 'stable')
                vers[Spec.STATUS[ds]].append(ev)
                vmap[ev] = (dist['dist'], dv)

        # for each rel status only take the max one.
        for i in range(len(vers)):
            vers[i] = vers[i] and max(vers[i]) or None

        ev = self._get_best_version(vers, spec, quiet=False)
        return vmap[ev]

    def _get_best_version(self, vers, spec, quiet):
        # Is there any result at the desired release status?
        want = [
            v
            for lvl, v in enumerate(vers)
            if lvl >= self.opts.status and v is not None
        ]
        if want:
            ver = max(want)
            if not quiet:
                logger.info(_("best version: %s %s"), spec.name, ver)
            return ver

        # Not found: is there any hint we can give?
        if self.opts.status > Spec.TESTING and vers[Spec.TESTING]:
            hint = (vers[Spec.TESTING], _('testing'))
        elif self.opts.status > Spec.UNSTABLE and vers[Spec.UNSTABLE]:
            hint = (vers[Spec.UNSTABLE], _('unstable'))
        else:
            hint = None

        msg = _("no suitable version found for %s") % spec
        if hint:
            msg += _(" but there is version %s at level %s") % hint

        raise ResourceNotFound(msg)

    def get_meta(self, spec):
        """
        Return the content of the ``META.json`` file for *spec*.

        Return the object obtained parsing the JSON.
        """
        if spec.is_name():
            # Get the metadata from the API
            try:
                data = self.api.dist(spec.name)
            except NotFound:
                # Distro not found: maybe it's an extension?
                ext = self.api.ext(spec.name)
                name, ver = self.get_best_version_from_ext(ext, spec)
                return self.api.meta(name, ver)
            else:
                ver = self.get_best_version(data, spec)
                return self.api.meta(spec.name, ver)

        elif spec.is_dir():
            # Get the metadata from a directory
            fn = os.path.join(spec.dirname, 'META.json')
            logger.debug("reading %s", fn)
            if not os.path.exists(fn):
                raise PgxnClientException(
                    _("file 'META.json' not found in '%s'") % spec.dirname
                )

            with open(fn) as f:
                return load_json(f)

        elif spec.is_file():
            arc = archive.from_spec(spec)
            return arc.get_meta()

        elif spec.is_url():
            with network.get_file(spec.url) as fin:
                with temp_dir() as dir:
                    fn = network.download(fin, dir)
                    arc = archive.from_file(fn)
                    return arc.get_meta()

        else:
            assert False


class WithSpecLocal(WithSpec):
    """
    Mixin to implement commands that can also refer to a local file or dir.
    """

    @classmethod
    def customize_parser(self, parser, subparsers, epilog=None, **kwargs):
        epilog = (
            _(
                """
SPEC may also be a local zip file or unpacked directory, but in this case
it should contain at least a '%s', for instance '.%spkgname.zip'.
"""
            )
            % (os.sep, os.sep)
            + (epilog or "")
        )

        subp = super(WithSpecLocal, self).customize_parser(
            parser, subparsers, epilog=epilog, **kwargs
        )

        return subp

    def get_spec(self, **kwargs):
        kwargs['_can_be_local'] = True
        return super(WithSpecLocal, self).get_spec(**kwargs)


class WithSpecUrl(WithSpec):
    """
    Mixin to implement commands that can also refer to a URL.
    """

    @classmethod
    def customize_parser(self, parser, subparsers, epilog=None, **kwargs):
        epilog = (
            _(
                """
SPEC may also be an url specifying a protocol such as 'http://' or 'https://'.
"""
            )
            + (epilog or "")
        )

        subp = super(WithSpecUrl, self).customize_parser(
            parser, subparsers, epilog=epilog, **kwargs
        )

        return subp

    def get_spec(self, **kwargs):
        kwargs['_can_be_url'] = True
        return super(WithSpecUrl, self).get_spec(**kwargs)


class WithPgConfig(object):
    """
    Mixin to implement commands that should query :program:`pg_config`.
    """

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        """
        Add the ``--pg_config`` option to the options parser.
        """
        subp = super(WithPgConfig, self).customize_parser(
            parser, subparsers, **kwargs
        )

        subp.add_argument(
            '--pg_config',
            metavar="PROG",
            default='pg_config',
            help=_(
                "the pg_config executable to find the database"
                " [default: %(default)s]"
            ),
        )

        return subp

    def call_pg_config(self, what, _cache={}):
        """
        Call :program:`pg_config` and return its output.
        """
        if what in _cache:
            return _cache[what]

        logger.debug("running pg_config --%s", what)
        cmdline = [self.get_pg_config(), "--%s" % what]
        p = self.popen(cmdline, stdout=PIPE)
        out, err = p.communicate()
        if p.returncode:
            raise ProcessError(
                _("command returned %s: %s") % (p.returncode, cmdline)
            )

        out = out.rstrip().decode('utf-8')
        rv = _cache[what] = out
        return rv

    def get_pg_config(self):
        """
        Return the absolute path of the pg_config binary.
        """
        pg_config = self.opts.pg_config
        if os.path.split(pg_config)[0]:
            pg_config = os.path.abspath(pg_config)
        else:
            pg_config = find_executable(pg_config)
        if not pg_config:
            raise PgxnClientException(_("pg_config executable not found"))
        return pg_config


class WithMake(WithPgConfig):
    """
    Mixin to implement commands that should invoke :program:`make`.
    """

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        """
        Add the ``--make`` option to the options parser.
        """
        subp = super(WithMake, self).customize_parser(
            parser, subparsers, **kwargs
        )

        subp.add_argument(
            '--make',
            metavar="PROG",
            default=self._find_default_make(),
            help=_(
                "the 'make' executable to use to build the extension "
                "[default: %(default)s]"
            ),
        )

        return subp

    def run_make(self, cmd, dir, env=None, sudo=None):
        """Invoke make with the selected command.

        :param cmd: the make target or list of options to pass make
        :param dir: the direcrory to run the command into
        :param env: variables to add to the make environment
        :param sudo: if set, use the provided command/arg to elevate
            privileges
        """
        # check if the directory contains a makefile
        for fn in ('GNUmakefile', 'makefile', 'Makefile'):
            if os.path.exists(os.path.join(dir, fn)):
                break
        else:
            raise PgxnClientException(
                _("no Makefile found in the extension root")
            )

        cmdline = []

        if sudo:
            cmdline.extend(shlex.split(sudo))

        cmdline.extend(
            [self.get_make(), 'PG_CONFIG=%s' % self.get_pg_config()]
        )

        if isinstance(cmd, six.string_types):
            cmdline.append(cmd)
        else:  # a list
            cmdline.extend(cmd)

        logger.debug(_("running: %s"), cmdline)
        p = self.popen(cmdline, cwd=dir, shell=False, env=env, close_fds=True)
        p.communicate()
        if p.returncode:
            raise ProcessError(
                _("command returned %s: %s")
                % (p.returncode, ' '.join(cmdline))
            )

    def get_make(self, _cache=[]):
        """
        Return the path of the make binary.
        """
        # the cache is not for performance but to return a consistent value
        # even if the cwd is changed
        if _cache:
            return _cache[0]

        make = self.opts.make

        if os.path.split(make)[0]:
            # At least a relative dir specified.
            if not os.path.exists(make):
                raise PgxnClientException(
                    _("make executable not found: %s") % make
                )

            # Convert to abs path to be robust in case the dir is changed.
            make = os.path.abspath(make)

        else:
            # we don't find make here and convert to abs path because it's a
            # security hole: make may be run under sudo and in this case we
            # don't want root to execute a make hacked in an user local dir
            if not find_executable(make):
                raise PgxnClientException(
                    _("make executable not found: %s") % make
                )

        _cache.append(make)
        return make

    @classmethod
    def _find_default_make(self):
        for make in ('gmake', 'make'):
            path = find_executable(make)
            if path:
                return make

        # if nothing was found, fall back on 'gmake'. If it was missing we
        # will give an error when attempting to use it
        return 'gmake'


class WithSudo(object):
    """
    Mixin to implement commands that may invoke sudo.
    """

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        subp = super(WithSudo, self).customize_parser(
            parser, subparsers, **kwargs
        )

        g = subp.add_mutually_exclusive_group()
        g.add_argument(
            '--sudo',
            metavar="PROG",
            const='sudo',
            nargs="?",
            help=_(
                "run PROG to elevate privileges when required"
                " [default: %(const)s]"
            ),
        )
        g.add_argument(
            '--nosudo',
            dest='sudo',
            action='store_false',
            help=_(
                "never elevate privileges "
                "(no more needed: for backward compatibility)"
            ),
        )

        return subp


class WithDatabase(object):
    """
    Mixin to implement commands that should communicate to a database.
    """

    @classmethod
    def customize_parser(self, parser, subparsers, epilog=None, **kwargs):
        """
        Add the options related to database connections.
        """
        epilog = (
            _(
                """
The default database connection options depend on the value of environment
variables PGDATABASE, PGHOST, PGPORT, PGUSER.
"""
            )
            + (epilog or "")
        )

        subp = super(WithDatabase, self).customize_parser(
            parser, subparsers, epilog=epilog, **kwargs
        )

        g = subp.add_argument_group(_("database connections options"))

        g.add_argument(
            '-d',
            '--dbname',
            metavar="DBNAME",
            help=_("database name to install into"),
        )
        g.add_argument(
            '-h',
            '--host',
            metavar="HOST",
            help=_("database server host or socket directory"),
        )
        g.add_argument(
            '-p',
            '--port',
            metavar="PORT",
            type=int,
            help=_("database server port"),
        )
        g.add_argument(
            '-U', '--username', metavar="NAME", help=_("database user name")
        )

        return subp

    def get_psql_options(self):
        """
        Return the cmdline options to connect to the specified database.
        """
        rv = []
        rv.append('-tAX')  # tuple only, unaligned, ignore psqlrc
        if self.opts.dbname:
            rv.extend(['--dbname', self.opts.dbname])
        if self.opts.host:
            rv.extend(['--host', self.opts.host])
        if self.opts.port:
            rv.extend(['--port', str(self.opts.port)])
        if self.opts.username:
            rv.extend(['--username', self.opts.username])
        return rv

    def get_psql_env(self):
        """
        Return a dict with env variables to connect to the specified db.
        """
        rv = {}
        if self.opts.dbname:
            rv['PGDATABASE'] = self.opts.dbname
        if self.opts.host:
            rv['PGHOST'] = self.opts.host
        if self.opts.port:
            rv['PGPORT'] = str(self.opts.port)
        if self.opts.username:
            rv['PGUSER'] = self.opts.username
        return rv



"""
pgxnclient -- installation/loading commands implementation
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import os
import re
import shutil
import difflib
import logging
import tempfile
from subprocess import PIPE

import six

from pgxnclient import SemVer
from pgxnclient import archive
from pgxnclient import network
from pgxnclient.i18n import _, N_
from pgxnclient.utils import sha1
from pgxnclient.errors import (
    BadChecksum,
    PgxnClientException,
    InsufficientPrivileges,
)
from pgxnclient.commands import Command, WithDatabase, WithMake, WithPgConfig
from pgxnclient.commands import WithSpecUrl, WithSpecLocal, WithSudo
from pgxnclient.utils.temp import temp_dir
from pgxnclient.utils.strings import Identifier

logger = logging.getLogger('pgxnclient.commands')


class Download(WithSpecUrl, Command):
    name = 'download'
    description = N_("download a distribution from the network")

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        subp = super(Download, self).customize_parser(
            parser, subparsers, **kwargs
        )
        subp.add_argument(
            '--target',
            metavar='PATH',
            default='.',
            help=_('Target directory and/or filename to save'),
        )

        return subp

    def run(self):
        spec = self.get_spec()
        assert not spec.is_local()

        if spec.is_url():
            return self._run_url(spec)

        data = self.get_meta(spec)

        try:
            chk = data['sha1']
        except KeyError:
            raise PgxnClientException(
                "sha1 missing from the distribution meta"
            )

        with self.api.download(data['name'], SemVer(data['version'])) as fin:
            fn = network.download(fin, self.opts.target)

        self.verify_checksum(fn, chk)
        return fn

    def _run_url(self, spec):
        with network.get_file(spec.url) as fin:
            fn = network.download(fin, self.opts.target)

        return fn

    def verify_checksum(self, fn, chk):
        """Verify that a downloaded file has the expected sha1."""
        sha = sha1()
        logger.debug(_("checking sha1 of '%s'"), fn)
        f = open(fn, "rb")
        try:
            while 1:
                data = f.read(8192)
                if not data:
                    break
                sha.update(data)
        finally:
            f.close()

        sha = sha.hexdigest()
        if sha != chk:
            os.unlink(fn)
            logger.error(_("file %s has sha1 %s instead of %s"), fn, sha, chk)
            raise BadChecksum(_("bad sha1 in downloaded file"))


class InstallUninstall(WithMake, WithSpecUrl, WithSpecLocal, Command):
    """
    Base class to implement the ``install`` and ``uninstall`` commands.
    """

    def run(self):
        with temp_dir() as dir:
            return self._run(dir)

    def _run(self, dir):
        spec = self.get_spec()
        if spec.is_dir():
            pdir = os.path.abspath(spec.dirname)
        elif spec.is_file():
            pdir = archive.from_file(spec.filename).unpack(dir)
        elif not spec.is_local():
            self.opts.target = dir
            fn = Download(self.opts).run()
            pdir = archive.from_file(fn).unpack(dir)
        else:
            assert False

        self.maybe_run_configure(pdir)

        self._inun(pdir)

    def _inun(self, pdir):
        """Run the specific command, implemented in the subclass."""
        raise NotImplementedError

    def maybe_run_configure(self, dir):
        fn = os.path.join(dir, 'configure')
        logger.debug("checking '%s'", fn)
        if not os.path.exists(fn):
            return

        logger.info(_("running configure"))
        p = self.popen(fn, cwd=dir)
        p.communicate()
        if p.returncode:
            raise PgxnClientException(
                _("configure failed with return code %s") % p.returncode
            )


class SudoInstallUninstall(WithSudo, InstallUninstall):
    """
    Installation commands base class supporting sudo operations.
    """

    def run(self):
        if not self.is_libdir_writable() and not self.opts.sudo:
            dir = self.call_pg_config('libdir')
            raise InsufficientPrivileges(
                _(
                    "PostgreSQL library directory (%s) not writable: "
                    "you should run the program as superuser, or specify "
                    "a 'sudo' program"
                )
                % dir
            )

        return super(SudoInstallUninstall, self).run()

    def get_sudo_prog(self):
        if self.is_libdir_writable():
            return None  # not needed

        return self.opts.sudo

    def is_libdir_writable(self):
        """
        Check if the Postgres installation directory is writable.

        If it is, we will assume that sudo is not required to
        install/uninstall the library, so the sudo program will not be invoked
        or its specification will not be required.
        """
        dir = self.call_pg_config('libdir')
        logger.debug("testing if %s is writable", dir)
        try:
            f = tempfile.TemporaryFile(prefix="pgxn-", suffix=".test", dir=dir)
            f.write(b'test')
            f.close()
        except (IOError, OSError):
            rv = False
        else:
            rv = True

        return rv


class Install(SudoInstallUninstall):
    name = 'install'
    description = N_("download, build and install a distribution")

    def _inun(self, pdir):
        logger.info(_("building extension"))
        self.run_make('all', dir=pdir)

        logger.info(_("installing extension"))
        self.run_make('install', dir=pdir, sudo=self.get_sudo_prog())


class Uninstall(SudoInstallUninstall):
    name = 'uninstall'
    description = N_("remove a distribution from the system")

    def _inun(self, pdir):
        logger.info(_("removing extension"))
        self.run_make('uninstall', dir=pdir, sudo=self.get_sudo_prog())


class Check(WithDatabase, InstallUninstall):
    name = 'check'
    description = N_("run a distribution's test")

    def _inun(self, pdir):
        logger.info(_("checking extension"))
        upenv = self.get_psql_env()
        logger.debug("additional env: %s", upenv)
        env = os.environ.copy()
        env.update(upenv)

        cmd = ['installcheck']
        if 'PGDATABASE' in upenv:
            cmd.append("CONTRIB_TESTDB=" + env['PGDATABASE'])

        try:
            self.run_make(cmd, dir=pdir, env=env)
        except PgxnClientException:
            # if the test failed, copy locally the regression result
            for ext in ('out', 'diffs'):
                fn = os.path.join(pdir, 'regression.' + ext)
                if os.path.exists(fn):
                    dest = './regression.' + ext
                    if not os.path.exists(dest) or not os.path.samefile(
                        fn, dest
                    ):
                        logger.info(_('copying regression.%s'), ext)
                        shutil.copy(fn, dest)
            raise


class LoadUnload(
    WithPgConfig, WithDatabase, WithSpecUrl, WithSpecLocal, Command
):
    """
    Base class to implement the ``load`` and ``unload`` commands.
    """

    @classmethod
    def customize_parser(self, parser, subparsers, **kwargs):
        subp = super(LoadUnload, self).customize_parser(
            parser, subparsers, **kwargs
        )

        subp.add_argument(
            '--schema',
            metavar="SCHEMA",
            type=Identifier.parse_arg,
            help=_("use SCHEMA instead of the default schema"),
        )

        subp.add_argument(
            'extensions',
            metavar='EXT',
            nargs='*',
            help=_("only specified extensions [default: all]"),
        )

        return subp

    def get_pg_version(self):
        """Return the version of the selected database."""
        data = self.call_psql('SHOW server_version_num')
        pgver = self.parse_pg_version(data)
        logger.debug("PostgreSQL version: %s", '.'.join(map(str, pgver)))
        return pgver

    def parse_pg_version(self, data):
        data = data.rstrip()
        try:
            nums = (int(data[:-4]), int(data[-4:-2]), int(data[-2:]))
        except Exception:
            raise PgxnClientException(
                "cannot parse version number from '%s'" % data
            )

        if nums[0] >= 10:
            if nums[1]:
                raise PgxnClientException("weird version number: '%s'" % data)
            nums = (nums[0], nums[2])

        return nums

    def is_extension(self, name):
        fn = os.path.join(
            self.call_pg_config('sharedir'), "extension", name + ".control"
        )
        logger.debug("checking if exists %s", fn)
        return os.path.exists(fn)

    def call_psql(self, command):
        cmdline = [self.find_psql()]
        cmdline.extend(self.get_psql_options())
        if command is not None:
            cmdline.extend(['-c', command])

        logger.debug("calling %s", cmdline)
        p = self.popen(cmdline, stdout=PIPE)
        out, err = p.communicate()
        if p.returncode:
            raise PgxnClientException(
                "psql returned %s running command" % (p.returncode)
            )

        return out.decode('utf-8')

    def load_sql(self, filename=None, data=None):
        cmdline = [self.find_psql()]
        cmdline.extend(self.get_psql_options())
        # load via pipe to enable psql commands in the file
        if not data:
            logger.debug("loading sql from %s", filename)
            with open(filename, 'r') as fin:
                p = self.popen(cmdline, stdin=fin)
                p.communicate()
        else:
            if len(data) > 105:
                tdata = data[:100] + "..."
            else:
                tdata = data
            logger.debug('running sql command: "%s"', tdata)
            p = self.popen(cmdline, stdin=PIPE)
            # for Python 3: just assume default encoding will do
            if isinstance(data, six.text_type):
                data = data.encode()
            p.communicate(data)

        if p.returncode:
            raise PgxnClientException(
                "psql returned %s loading extension" % (p.returncode)
            )

    def find_psql(self):
        return self.call_pg_config('bindir') + '/psql'

    def find_sql_file(self, name, sqlfile):
        # In the extension the sql can be specified with a directory,
        # butit gets flattened into the target dir by the Makefile
        sqlfile = os.path.basename(sqlfile)

        sharedir = self.call_pg_config('sharedir')
        # TODO: we only check in contrib and in <name>: actually it may be
        # somewhere else - only the makefile knows!
        tries = [
            name + '/' + sqlfile,
            sqlfile.rsplit('.', 1)[0] + '/' + sqlfile,
            'contrib/' + sqlfile,
        ]
        tried = set()
        for fn in tries:
            if fn in tried:
                continue
            tried.add(fn)
            fn = sharedir + '/' + fn
            logger.debug("checking sql file in %s" % fn)
            if os.path.exists(fn):
                return fn
        else:
            raise PgxnClientException(
                "cannot find sql file for extension '%s': '%s'"
                % (name, sqlfile)
            )

    def patch_for_schema(self, fn):
        """
        Patch a sql file to set the schema where the commands are executed.

        If no schema has been requested, return the data unchanged.
        Else, ask for confirmation and return the data for a patched file.

        The schema is only useful for PG < 9.1: for proper PG extensions there
        is no need to patch the sql.
        """
        schema = self.opts.schema

        f = open(fn)
        try:
            data = f.read()
        finally:
            f.close()

        if not schema:
            return data

        self._check_schema_exists(schema)

        re_path = re.compile(r'SET\s+search_path\s*(?:=|to)\s*([^;]+);', re.I)
        m = re_path.search(data)
        if m is None:
            newdata = ("SET search_path = %s;\n\n" % schema) + data
        else:
            newdata = re_path.sub("SET search_path = %s;" % schema, data)

        diff = ''.join(
            difflib.unified_diff(
                [r + '\n' for r in data.splitlines()],
                [r + '\n' for r in newdata.splitlines()],
                fn,
                fn + ".schema",
            )
        )
        msg = _(
            """
In order to operate in the schema %s, the following changes will be
performed:\n\n%s\n\nDo you want to continue?"""
        )
        self.confirm(msg % (schema, diff))

        return newdata

    def _register_loaded(self, fn):
        if not hasattr(self, '_loaded'):
            self._loaded = []

        self._loaded.append(fn)

    def _is_loaded(self, fn):
        return hasattr(self, '_loaded') and fn in self._loaded

    def _check_schema_exists(self, schema):
        cmdline = [self.find_psql()]
        cmdline.extend(self.get_psql_options())
        cmdline.extend(['-c', 'SET search_path=%s' % schema])
        p = self.popen(cmdline, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        p.communicate()
        if p.returncode:
            raise PgxnClientException("schema %s does not exist" % schema)

    def _get_extensions(self):
        """
        Return a list of pairs (name, sql file) to be loaded/unloaded.

        Items are in loading order.
        """
        spec = self.get_spec()
        dist = self.get_meta(spec)

        if 'provides' not in dist:
            # No 'provides' specified: assume a single extension named
            # after the distribution. This is automatically done by PGXN,
            # but we should do ourselves to deal with local META files
            # not mangled by the PGXN upload script yet.
            name = dist['name']
            for ext in self.opts.extensions:
                if ext != name:
                    raise PgxnClientException(
                        "can't find extension '%s' in the distribution '%s'"
                        % (name, spec)
                    )

            return [(name, None)]

        rv = []

        if not self.opts.extensions:
            # All the extensions, in the order specified
            # (assume we got an orddict from json)
            for name, data in dist['provides'].items():
                rv.append((name, data.get('file')))
        else:
            # Only the specified extensions
            for name in self.opts.extensions:
                try:
                    data = dist['provides'][name]
                except KeyError:
                    raise PgxnClientException(
                        "can't find extension '%s' in the distribution '%s'"
                        % (name, spec)
                    )
                rv.append((name, data.get('file')))

        return rv


class Load(LoadUnload):
    name = 'load'
    description = N_("load a distribution's extensions into a database")

    def run(self):
        items = self._get_extensions()
        for (name, sql) in items:
            self.load_ext(name, sql)

    def load_ext(self, name, sqlfile):
        logger.debug(_("loading extension '%s' with file: %s"), name, sqlfile)

        if sqlfile and not sqlfile.endswith('.sql'):
            logger.info(
                _(
                    "the specified file '%s' doesn't seem SQL:"
                    " assuming '%s' is not a PostgreSQL extension"
                ),
                sqlfile,
                name,
            )
            return

        pgver = self.get_pg_version()

        if pgver >= (9, 1, 0):
            if self.is_extension(name):
                self.create_extension(name)
                return
            else:
                self.confirm(
                    _(
                        """\
The extension '%s' doesn't contain a control file:
it will be installed as a loose set of objects.
Do you want to continue?"""
                    )
                    % name
                )

        confirm = False
        if not sqlfile:
            sqlfile = name + '.sql'
            confirm = True

        fn = self.find_sql_file(name, sqlfile)
        if confirm:
            self.confirm(
                _(
                    """\
The extension '%s' doesn't specify a SQL file.
'%s' is probably the right one.
Do you want to load it?"""
                )
                % (name, fn)
            )

        # TODO: is confirmation asked only once? Also, check for repetition
        # in unload.
        if self._is_loaded(fn):
            logger.info(_("file %s already loaded"), fn)
        else:
            data = self.patch_for_schema(fn)
            self.load_sql(data=data)
            self._register_loaded(fn)

    def create_extension(self, name):
        name = Identifier(name)
        schema = self.opts.schema
        cmd = ["CREATE EXTENSION", name]
        if schema:
            cmd.extend(["SCHEMA", schema])

        cmd = " ".join(cmd) + ';'
        self.load_sql(data=cmd)


class Unload(LoadUnload):
    name = 'unload'
    description = N_("unload a distribution's extensions from a database")

    def run(self):
        items = self._get_extensions()

        if not self.opts.extensions:
            items.reverse()

        for (name, sql) in items:
            self.unload_ext(name, sql)

    def unload_ext(self, name, sqlfile):
        logger.debug(
            _("unloading extension '%s' with file: %s"), name, sqlfile
        )

        if sqlfile and not sqlfile.endswith('.sql'):
            logger.info(
                _(
                    "the specified file '%s' doesn't seem SQL:"
                    " assuming '%s' is not a PostgreSQL extension"
                ),
                sqlfile,
                name,
            )
            return

        pgver = self.get_pg_version()

        if pgver >= (9, 1, 0):
            if self.is_extension(name):
                self.drop_extension(name)
                return
            else:
                self.confirm(
                    _(
                        """\
The extension '%s' doesn't contain a control file:
will look for an SQL script to unload the objects.
Do you want to continue?"""
                    )
                    % name
                )

        if not sqlfile:
            sqlfile = name + '.sql'

        tmp = os.path.split(sqlfile)
        sqlfile = os.path.join(tmp[0], 'uninstall_' + tmp[1])

        fn = self.find_sql_file(name, sqlfile)
        self.confirm(
            _(
                """\
In order to unload the extension '%s' looks like you will have
to load the file '%s'.
Do you want to execute it?"""
            )
            % (name, fn)
        )

        data = self.patch_for_schema(fn)
        self.load_sql(data=data)

    def drop_extension(self, name):
        # TODO: cascade
        cmd = "DROP EXTENSION %s;" % Identifier(name)
        self.load_sql(data=cmd)



"""
pgxnclient -- temp files utilities
"""

# Copyright (C) 2011-2019 Daniele Varrazzo

# This file is part of the PGXN client

import shutil
import tempfile
import contextlib


@contextlib.contextmanager
def temp_dir():
    """Context manager to create a temp dir and delete after usage."""
    dir = tempfile.mkdtemp()
    yield dir
    shutil.rmtree(dir)
